# -*- coding: utf-8 -*-
#
#  Copyright 2020-2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from traceback import format_exc
from .mol import MOLRead
from ...exceptions import EmptyMolecule


class RXNRead:
    def __init__(self, line, ignore=False, log_buffer=None):
        self.__reactants_count = int(line[:3])
        self.__products_count = int(line[3:6]) + self.__reactants_count
        self.__reagents_count = int(line[6:].rstrip() or 0) + self.__products_count
        self.__molecules = []
        self.__ignore = ignore
        self.__errors = []
        if log_buffer is None:
            log_buffer = []
        self.__log_buffer = log_buffer

    def __call__(self, line):
        if self.__parser:
            try:
                x = self.__parser(line)
            except ValueError:
                if not self.__ignore:
                    raise
                if (lm := len(self.__molecules)) < self.__reactants_count:
                    self.__reactants_count -= 1
                    self.__products_count -= 1
                    self.__reagents_count -= 1
                    m = 'reactants'
                elif lm < self.__products_count:
                    self.__products_count -= 1
                    self.__reagents_count -= 1
                    m = 'products'
                else:
                    self.__reagents_count -= 1
                    m = 'reagents'

                self.__errors.append({'chytorch_molecule_role': m})
                self.__log_buffer.append(format_exc())
                self.__log_buffer.append('bad molecule ignored')
                self.__parser = None
                if lm == self.__reagents_count:  # empty molecule is last in list
                    self.__rend = True
                    return True
                self.__empty_skip = True
            else:
                if x:
                    self.__im = 4
                    mol = self.__parser.getvalue()
                    if self.__title:
                        mol['title'] = self.__title
                    self.__molecules.append(mol)
                    self.__parser = None
                    if len(self.__molecules) == self.__reagents_count:
                        self.__rend = True
                        return True
        elif self.__empty_skip:
            if not line.startswith('$MOL'):
                return
            self.__empty_skip = False
            self.__im = 3
        elif self.__rend:
            raise ValueError('parser closed')
        elif self.__im == 4:
            if not line.startswith('$MOL'):
                raise ValueError('invalid RXN')
            self.__im = 3
        elif self.__im:
            if self.__im == 3:
                self.__title = line.strip()
            self.__im -= 1
        else:
            try:
                self.__parser = MOLRead(line, self.__log_buffer)
            except ValueError as e:
                if not self.__ignore:
                    raise
                if (lm := len(self.__molecules)) < self.__reactants_count:
                    self.__reactants_count -= 1
                    self.__products_count -= 1
                    self.__reagents_count -= 1
                    m = 'reactants'
                elif lm < self.__products_count:
                    self.__products_count -= 1
                    self.__reagents_count -= 1
                    m = 'products'
                else:
                    self.__reagents_count -= 1
                    m = 'reagents'

                if not isinstance(e, EmptyMolecule):
                    self.__errors.append({'chytorch_molecule_role': m})
                    self.__log_buffer.append(format_exc())
                    self.__log_buffer.append('bad molecule ignored')
                else:
                    self.__log_buffer.append('empty molecule ignored')

                if lm == self.__reagents_count:  # empty molecule is last in list
                    self.__rend = True
                    return True
                self.__empty_skip = True

    def getvalue(self):
        if self.__rend:
            return {'reactants': self.__molecules[:self.__reactants_count],
                    'products': self.__molecules[self.__reactants_count:self.__products_count],
                    'reagents': self.__molecules[self.__products_count:self.__reagents_count],
                    'meta': {}, 'errors': self.__errors}
        raise ValueError('reaction not complete')

    __parser = None
    __empty_skip = __rend = False
    __im = 4


__all__ = ['RXNRead']



# -*- coding: utf-8 -*-
#
#  Copyright 2021, 2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from typing import List, TYPE_CHECKING, Union
from ...exceptions import ValenceError


if TYPE_CHECKING:
    from chython import MoleculeContainer


class Resonance:
    __slots__ = ()

    def fix_resonance(self: Union['MoleculeContainer', 'Resonance'], *, logging=False,
                      _fix_stereo=True) -> Union[bool, List[int]]:
        """
        Transform biradical or dipole resonance structures into neutral form. Return True if structure form changed.

        :param logging: return list of changed atoms.
        """
        atoms = self._atoms
        charges = self._charges
        radicals = self._radicals
        bonds = self._bonds
        calc_implicit = self._calc_implicit
        entries, exits, rads, constrains, nitrogen_cat, nitrogen_ani, sulfur_cat = self.__entries()
        hs = set()
        while len(rads) > 1:
            n = rads.pop()
            for path in self.__find_delocalize_path(n, rads, constrains, True):
                radicals[n] = False
                hs.add(n)
                for n, m, b in path:
                    hs.add(m)
                    bonds[n][m]._Bond__order = b  # noqa
                radicals[m] = False  # noqa
                rads.discard(m)
                break  # path found
            # path not found. atom n keep as is
        while entries and exits:
            n = entries.pop()
            for path in self.__find_delocalize_path(n, exits, constrains, False):
                l, m, b = path[-1]
                if n in nitrogen_cat and m in nitrogen_ani:
                    continue

                c_m = charges[m] - 1
                if m in sulfur_cat:  # prevent X-[S+]=X >> X=S=X
                    if b != 1:
                        continue
                else:  # check cations end valence.
                    try:
                        atoms[m].valence_rules(c_m, radicals[m], sum(int(y) for x, y in bonds[m].items() if x != l) + b)
                    except ValenceError:
                        continue

                charges[n] += 1
                hs.add(n)
                for n, m, b in path:
                    hs.add(m)
                    bonds[n][m]._Bond__order = b  # noqa
                charges[m] = c_m
                exits.discard(m)
                break  # path from negative atom to positive atom found.
            # path not found. keep negative atom n as is
        if hs:
            for n in hs:
                calc_implicit(n)
            self.flush_cache()
            if _fix_stereo:
                self.fix_stereo()
            if logging:
                return list(hs)
            return True
        elif logging:
            return []
        return False

    def __find_delocalize_path(self: 'MoleculeContainer', start, finish, constrains, odd_only):
        bonds = self._bonds
        stack = [(start, n, 0, b.order + 1) for n, b in bonds[start].items() if n in constrains and b.order < 3]
        path = []
        seen = {start}
        while stack:
            last, current, depth, order = stack.pop()
            if len(path) > depth:
                seen.difference_update(x for _, x, _ in path[depth:])
                path = path[:depth]

            path.append((last, current, order))

            if current in finish:
                if odd_only:  # radicals!
                    if len(path) % 2:
                        yield path
                    else:  # invalid path
                        continue
                elif depth:  # one bonded ignored. we search double bond transfer! A=A-A >> A-A=A.
                    yield path

            depth += 1
            seen.add(current)
            diff = -1 if depth % 2 else 1
            stack.extend((current, n, depth, bo) for n, b in bonds[current].items()
                         if n not in seen and n in constrains and 1 <= (bo := b.order + diff) <= 3)

    def __entries(self: 'MoleculeContainer'):
        hybridization = self.hybridization
        neighbors = self.neighbors
        charges = self._charges
        radicals = self._radicals
        bonds = self._bonds
        atoms = self._atoms
        errors = {n for n, h in self._hydrogens.items() if h is None}

        transfer = set()
        entries = set()
        exits = set()
        rads = set()
        nitrogen_cat = set()
        nitrogen_ani = set()
        sulfur_cat = set()
        for n, a in atoms.items():
            if a.atomic_number not in {5, 6, 7, 8, 14, 15, 16, 33, 34, 52}:
                # filter non-organic set, halogens and aromatics
                continue
            elif radicals[n]:
                rads.add(n)
            elif charges[n] == -1:
                if (lb := len(bonds[n])) == 4 and a.atomic_number == 5:  # skip boron
                    continue
                elif lb == 6 and a.atomic_number == 15:  # skip [P-]X6
                    continue
                if n in errors:  # only valid anions accepted
                    continue
                entries.add(n)
            elif charges[n] == 1:
                lb = len(bonds[n])
                if a.atomic_number == 7:
                    if lb == 4:  # skip ammonia
                        continue
                    elif lb == 2 and hybridization(n) == 3:  # skip Azide
                        (n1, b1), (n2, b2) = bonds[n].items()
                        if b1.order == b2.order == 2 and (charges[n1] == -1 and atoms[n1].atomic_number == 7 or
                                                          charges[n2] == -1 and atoms[n2].atomic_number == 7):
                            continue
                    elif lb == 3 and hybridization(n) == 2:  # X=[N+](-X)-X - prevent N-N migration
                        nitrogen_ani.add(n)
                elif a.atomic_number == 15 and lb == 4:  # skip [P+]R4
                    continue
                elif a.atomic_number == 16:
                    if lb == 2 and hybridization(n) == 2:  # ad-hoc for X-[S+]=X
                        sulfur_cat.add(n)
                    elif lb == 3 and hybridization(n) == 1:  # ad-hoc for X-[S+](-X)-X
                        continue
                exits.add(n)
            transfer.add(n)

        if exits or entries:  # try to move cation to nitrogen. saturation fixup.
            for n, a in self._atoms.items():
                if a.atomic_number == 7 and not charges[n]:
                    if hybridization(n) == 1 and neighbors(n) <= 3:  # any amine - potential e-donor
                        entries.add(n)
                        nitrogen_cat.add(n)
                    elif hybridization(n) == 3 and neighbors(n) == 1:  # N#X-[X-] >> [N-]=X=X
                        exits.add(n)
                        nitrogen_ani.add(n)
        return entries, exits, rads, transfer, nitrogen_cat, nitrogen_ani, sulfur_cat


__all__ = ['Resonance']



# -*- coding: utf-8 -*-
#
#  Copyright 2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from asyncio import new_event_loop
from os.path import join
from tempfile import TemporaryDirectory


loop = browser = None


async def render(s, t, width, height, scale):
    page = await browser.newPage()
    await page.setViewport({'deviceScaleFactor': scale, 'width': width, 'height': height})
    await page.goto(f'file://{s}')
    element = await page.querySelector('svg')
    await element.screenshot({'path': t})
    await page.close()


def svg2png(svg: str, width: int = 1000, height: int = 1000, scale: float = 10.):
    global loop, browser

    if loop is None:  # lazy browser launcher
        from pyppeteer import launch

        loop = new_event_loop()
        browser = loop.run_until_complete(launch())
    elif browser is None:
        raise ImportError('pyppeteer initialization failed')

    with TemporaryDirectory() as tmpdir:
        with open(s := join(tmpdir, 'input.svg'), 'w') as f:
            f.write(svg)

        loop.run_until_complete(render(s, (t := join(tmpdir, 'output.png')), width, height, scale))

        with open(t, 'rb') as f:
            return f.read()


__all__ = ['svg2png']



# -*- coding: utf-8 -*-
#
#  Copyright 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from io import StringIO, TextIOWrapper
from pathlib import Path


class _MDLWrite:
    def __init__(self, file, *, mapping: bool = True, append: bool = False):
        """
        :param mapping: write atom mapping.
        """
        self._mapping = mapping

        if isinstance(file, str):
            self._file = open(file, 'a' if append else 'w')
            self._is_buffer = False
        elif isinstance(file, Path):
            self._file = file.open('a' if append else 'w')
            self._is_buffer = False
        elif isinstance(file, (TextIOWrapper, StringIO)):
            self._file = file
            self._is_buffer = True
        else:
            raise TypeError('invalid file. '
                            'TextIOWrapper, StringIO, BytesIO, BufferedReader and BufferedIOBase subclasses possible')

    def close(self, force=False):
        """
        close opened file

        :param force: force closing of externally opened file or buffer
        """
        self.write = self.__write_closed

        if not self._is_buffer or force:
            self._file.close()

    def __enter__(self):
        return self

    def __exit__(self, _type, value, traceback):
        self.close()

    @staticmethod
    def __write_closed(_):
        raise ValueError('I/O operation on closed writer')


__all__ = ['_MDLWrite']



# -*- coding: utf-8 -*-
#
#  Copyright 2021, 2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
cimport cython
from cpython.mem cimport PyMem_Malloc, PyMem_Free
from libc.math cimport ldexp

from chython.containers.bonds import Bond

# Format specification::
#
# Big endian bytes order
# 8 bit - 0x02 (current format specification)
# 12 bit - number of atoms
# 12 bit - cis/trans stereo block size
#
# Atom block 9 bytes (repeated):
# 12 bit - atom number
# 4 bit - number of neighbors
# 2 bit tetrahedron sign (00 - not stereo, 10 or 11 - has stereo)
# 2 bit - allene sign
# 5 bit - isotope (00000 - not specified, over = isotope - common_isotope + 16)
# 7 bit - atomic number (<=118)
# 32 bit - XY float16 coordinates
# 3 bit - hydrogens (0-7). Note: 7 == None
# 4 bit - charge (charge + 4. possible range -4 - 4)
# 1 bit - radical state
# Connection table: flatten list of neighbors. neighbors count stored in atom block.
# For example CC(=O)O - {1: [2], 2: [1, 3, 4], 3: [2], 4: [2]} >> [2, 1, 3, 4, 2, 2].
# Repeated block (equal to bonds count).
# 24 bit - paired 12 bit numbers.
# Bonds order block 3 bit per bond zero-padded to full byte at the end.
# Cis/trans data block (repeated):
# 24 bit - atoms pair
# 7 bit - zero padding. in future can be used for extra bond-level stereo, like atropoisomers.
# 1 bit - sign

@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
def unpack(const unsigned char[::1] data not None):
    cdef char *charges
    cdef unsigned char a, b, c, d, isotope, atomic_number, neighbors_count, s = 0, nc, version
    cdef unsigned char *atoms, *hydrogens, *neighbors, *radicals, *orders, *is_tet, *is_all
    cdef bint *stereo_sign, *ct_sign
    cdef unsigned short atoms_count, bonds_count = 0, cis_trans_count, order_count
    cdef unsigned short i, j, k = 0, n, m, buffer_b, shift = 0
    cdef unsigned short *mapping, *isotopes, *cis_trans_1, *cis_trans_2, *connections
    cdef unsigned int size, atoms_shift = 4, bonds_shift, order_shift, cis_trans_shift
    cdef double *x_coord, *y_coord
    cdef unsigned char[4096] seen

    cdef object bond, py_n, py_m
    cdef dict py_charges, py_radicals, py_hydrogens, py_plane, py_bonds, py_ngb
    cdef dict py_atoms_stereo, py_allenes_stereo, py_cis_trans_stereo
    cdef list py_mapping, py_atoms, py_isotopes, py_bonds_flat

    # read header
    version = data[0]
    a, b, c = data[1], data[2], data[3]
    atoms_count = a << 4| b >> 4
    cis_trans_count = (b & 0x0f) << 8 | c

    # allocate memory
    charges = <char*> PyMem_Malloc(atoms_count * sizeof(char))
    radicals = <unsigned char*> PyMem_Malloc(atoms_count * sizeof(unsigned char))
    atoms = <unsigned char*> PyMem_Malloc(atoms_count * sizeof(unsigned char))
    hydrogens = <unsigned char*> PyMem_Malloc(atoms_count * sizeof(unsigned char))
    neighbors = <unsigned char*> PyMem_Malloc(atoms_count * sizeof(unsigned char))
    is_tet = <unsigned char*> PyMem_Malloc(atoms_count * sizeof(unsigned char))
    is_all = <unsigned char*> PyMem_Malloc(atoms_count * sizeof(unsigned char))
    stereo_sign = <bint*> PyMem_Malloc(atoms_count * sizeof(bint))
    mapping = <unsigned short*> PyMem_Malloc(atoms_count * sizeof(unsigned short))
    isotopes = <unsigned short*> PyMem_Malloc(atoms_count * sizeof(unsigned short))
    x_coord = <double*> PyMem_Malloc(atoms_count * sizeof(double))
    y_coord = <double*> PyMem_Malloc(atoms_count * sizeof(double))

    if not charges or not radicals or not atoms or not hydrogens or not neighbors or not is_tet or not is_all:
        raise MemoryError()
    if not stereo_sign or not mapping or not isotopes or not x_coord or not y_coord:
        raise MemoryError()

    # unpack atom block to separate attributes arrays
    for i in range(atoms_count):
        a, b = data[atoms_shift], data[atoms_shift + 1]
        mapping[i] = n = a << 4 | b >> 4
        seen[n] = 0  # erase random value
        neighbors[i] = neighbors_count = b & 0x0f
        bonds_count += neighbors_count

        a, b = data[atoms_shift + 2], data[atoms_shift + 3]
        if a >> 7:  # tetrahedron bit set
            is_tet[i] = 1
            is_all[i] = 0
            stereo_sign[i] = a & 0x40  # mask th bit
        else:
            is_tet[i] = 0
            if a >> 5:  # allene bit set
                is_all[i] = 1
                stereo_sign[i] = a & 0x10  # mask al bit
            else:
                is_all[i] = 0

        atoms[i] = atomic_number = b & 0x7f
        isotope = (a & 0x0f) << 1 | b >> 7
        if isotope:
            isotopes[i] = common_isotopes[atomic_number] + isotope
        else:
            isotopes[i] = 0

        a, b = data[atoms_shift + 4], data[atoms_shift + 5]
        x_coord[i] = double_from_bytes(a, b)
        a, b = data[atoms_shift + 6], data[atoms_shift + 7]
        y_coord[i] = double_from_bytes(a, b)

        a = data[atoms_shift + 8]
        hydrogens[i] = a >> 5
        charges[i] = ((a >> 1) & 0x0f) - 4
        radicals[i] = a & 0x01
        atoms_shift += 9

    # calculate bonds count and pack sections
    bonds_count /= 2

    if version == 2:
        order_count = bonds_count * 3
        if order_count % 8:
            order_count = order_count / 8 + 1
        else:
            order_count /= 8
    elif version == 0:
        order_count = bonds_count / 5
        if bonds_count % 5:
            order_count += 1
        order_count *= 2

    bonds_shift = atoms_shift
    order_shift = bonds_shift + 3 * bonds_count
    cis_trans_shift = order_count + order_shift
    size = cis_trans_shift + 4 * cis_trans_count

    if bonds_count:
        # keep 4 extra cells for padding
        orders = <unsigned char*> PyMem_Malloc((bonds_count + 4) * sizeof(unsigned char))
        connections = <unsigned short*> PyMem_Malloc(2 * bonds_count * sizeof(unsigned short))

        # connection table is bidirected
        for i in range(0, 2 * bonds_count, 2):
            a, b, c = data[bonds_shift], data[bonds_shift + 1], data[bonds_shift + 2]
            connections[i] = a << 4| b >> 4
            connections[i + 1] = (b & 0x0f) << 8 | c
            bonds_shift += 3

        # collect flat bond order list
        i = 0
        if version == 2:
            for j in range(order_shift, cis_trans_shift):
                # 3 3 2 | 1 3 3 1 | 2 3 3
                a = data[j]
                if s == 1:
                    orders[i] = buffer_b | a >> 7
                    orders[i + 1] = (a >> 4) & 0x7
                    orders[i + 2] = (a >> 1) & 0x7
                    buffer_b = (a & 1) << 2
                    s = 2
                    i += 3
                elif s == 2:
                    orders[i] = buffer_b | a >> 6
                    orders[i + 1] = (a >> 3) & 0x7
                    orders[i + 2] = a & 0x7
                    i += 3
                    s = 0
                else:
                    orders[i] = a >> 5
                    orders[i + 1] = (a >> 2) & 0x7
                    buffer_b = (a & 0x3) << 1
                    s = 1
                    i += 2
        elif version == 0:
            for j in range(order_shift, cis_trans_shift, 2):
                # 0 3 3 1 | 2 3 3
                a, b = data[j], data[j + 1]
                orders[i] = a >> 4
                orders[i + 1] = (a >> 1) & 0x7
                orders[i + 2] = (a & 0x1) << 2 | b >> 6
                orders[i + 3] = (b >> 3) & 0x7
                orders[i + 4] = b & 0x7
                i += 5

        if cis_trans_count:
            cis_trans_1 = <unsigned short *> PyMem_Malloc(cis_trans_count * sizeof(unsigned short))
            cis_trans_2 = <unsigned short *> PyMem_Malloc(cis_trans_count * sizeof(unsigned short))
            ct_sign = <bint *> PyMem_Malloc(cis_trans_count * sizeof(bint))
            if not cis_trans_1 or not cis_trans_2 or not ct_sign:
                raise MemoryError()

            for i in range(cis_trans_count):
                a, b = data[cis_trans_shift], data[cis_trans_shift + 1]
                c, d = data[cis_trans_shift + 2], data[cis_trans_shift + 3]
                cis_trans_1[i] = a << 4 | b >> 4
                cis_trans_2[i] = (b & 0x0f) << 8 | c
                ct_sign[i] = d  # d = 0x01 or 0x00
                cis_trans_shift += 4

    # define returned data
    py_mapping = []
    py_atoms = []
    py_isotopes = []
    py_charges = {}
    py_radicals = {}
    py_hydrogens = {}
    py_plane = {}
    py_atoms_stereo = {}
    py_allenes_stereo = {}
    py_cis_trans_stereo = {}
    py_bonds = {}
    py_bonds_flat = []

    for i in range(atoms_count):
        n = mapping[i]
        py_n = n  # shared py int obj

        # fill intermediate data
        py_mapping.append(py_n)
        py_atoms.append(atoms[i])
        py_isotopes.append(isotopes[i] or None)

        py_charges[py_n] = charges[i]
        py_radicals[py_n] = bool(radicals[i])
        if hydrogens[i] == 7:
            py_hydrogens[py_n] = None
        else:
            py_hydrogens[py_n] = hydrogens[i]

        py_plane[py_n] = (x_coord[i], y_coord[i])

        if is_tet[i]:
            py_atoms_stereo[py_n] = stereo_sign[i]
        elif is_all[i]:
            py_allenes_stereo[py_n] = stereo_sign[i]

        py_bonds[py_n] = py_ngb = {}
        seen[n] = 1

        nc = neighbors[i]
        for j in range(shift, shift + nc):
            m = connections[j]
            py_m = m
            if seen[m]:  # bond partially exists. need back-connection.
                py_ngb[py_m] = py_bonds[py_m][py_n]
            else:
                bond = object.__new__(Bond)
                bond._Bond__order = orders[k] + 1
                bond._Bond__n = py_n
                bond._Bond__m = py_m
                py_ngb[py_m] = bond
                py_bonds_flat.append(bond)
                k += 1
        shift += nc

    for i in range(cis_trans_count):
        py_cis_trans_stereo[(cis_trans_1[i], cis_trans_2[i])] = ct_sign[i]

    PyMem_Free(charges)
    PyMem_Free(radicals)
    PyMem_Free(atoms)
    PyMem_Free(hydrogens)
    PyMem_Free(neighbors)
    PyMem_Free(is_tet)
    PyMem_Free(is_all)
    PyMem_Free(stereo_sign)
    PyMem_Free(mapping)
    PyMem_Free(isotopes)
    PyMem_Free(x_coord)
    PyMem_Free(y_coord)
    if bonds_count:
        PyMem_Free(connections)
        PyMem_Free(orders)
        if cis_trans_count:
            PyMem_Free(cis_trans_1)
            PyMem_Free(cis_trans_2)
            PyMem_Free(ct_sign)
    return (py_mapping, py_atoms, py_isotopes,
            py_charges, py_radicals, py_hydrogens, py_plane, py_bonds,
            py_atoms_stereo, py_allenes_stereo, py_cis_trans_stereo, size, py_bonds_flat)


cdef short[119] common_isotopes
common_isotopes[:] = [0, -15, -12, -9, -7, -5, -4, -2, 0, 3, 4, 7, 8, 11, 12, 15, 16, 19, 24, 23, 24, 29,
                      32, 35, 36, 39, 40, 43, 43, 48, 49, 54, 57, 59, 63, 64, 68, 69, 72, 73, 75, 77,
                      80, 82, 85, 87, 90, 92, 96, 99, 103, 106, 112, 111, 115, 117, 121, 123, 124, 125,
                      128, 129, 134, 136, 141, 143, 147, 149, 151, 153, 157, 159, 162, 165, 168, 170,
                      174, 176, 179, 181, 185, 188, 191, 193, 193, 194, 206, 207, 210, 211, 216, 215,
                      222, 221, 228, 227, 231, 231, 235, 236, 241, 242, 243, 244, 245, 254, 253, 254,
                      254, 262, 265, 265, 269, 262, 273, 273, 277, 281, 278]


cdef double double_from_bytes(unsigned char a, unsigned char b):
    cdef bint sign
    cdef int e
    cdef unsigned int f
    cdef double x

    sign = a >> 7
    e = (a >> 2) & 0x1f
    f = ((a & 0x03) << 8) | b

    x = f / 1024.
    if e:
        x += 1.
        e -= 15
    else:
        e = -14

    x = ldexp(x, e)
    if sign:
        return -x
    return x



# -*- coding: utf-8 -*-
#
#  Copyright 2018-2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from array import array
from functools import cached_property
from itertools import chain, product
from typing import Dict, List, Tuple, Union
from . import molecule  # cyclic imports resolve
from .bonds import Bond, QueryBond
from .graph import Graph
from ..algorithms.smiles import QuerySmiles
from ..algorithms.stereo import Stereo
from ..periodictable import AnyElement, AnyMetal, Element, ListElement, QueryElement
from ..periodictable.element import Query


def _validate_neighbors(neighbors):
    if neighbors is None:
        neighbors = ()
    elif isinstance(neighbors, int):
        if neighbors < 0 or neighbors > 14:
            raise ValueError('neighbors should be in range [0, 14]')
        neighbors = (neighbors,)
    elif isinstance(neighbors, (tuple, list)):
        if not all(isinstance(n, int) for n in neighbors):
            raise TypeError('neighbors should be list or tuple of ints')
        if any(n < 0 or n > 14 for n in neighbors):
            raise ValueError('neighbors should be in range [0, 14]')
        if len(set(neighbors)) != len(neighbors):
            raise ValueError('neighbors should be unique')
        neighbors = tuple(sorted(neighbors))
    else:
        raise TypeError('neighbors should be int or list or tuple of ints')
    return neighbors


class QueryContainer(Stereo, Graph[Query, QueryBond], QuerySmiles):
    __slots__ = ('_neighbors', '_hybridizations', '_atoms_stereo', '_cis_trans_stereo', '_allenes_stereo',
                 '_hydrogens', '_rings_sizes', '_heteroatoms', '_masked')

    _neighbors: Dict[int, Tuple[int, ...]]
    _hybridizations: Dict[int, Tuple[int, ...]]
    _atoms_stereo: Dict[int, bool]
    _allenes_stereo: Dict[int, bool]
    _cis_trans_stereo: Dict[Tuple[int, int], bool]
    _hydrogens: Dict[int, Tuple[int, ...]]
    _rings_sizes: Dict[int, Tuple[int, ...]]
    _heteroatoms: Dict[int, Tuple[int, ...]]
    _masked: Dict[int, bool]

    def __init__(self):
        super().__init__()
        self._neighbors = {}
        self._hybridizations = {}
        self._atoms_stereo = {}
        self._allenes_stereo = {}
        self._cis_trans_stereo = {}
        self._hydrogens = {}
        self._rings_sizes = {}
        self._heteroatoms = {}
        self._masked = {}

    def add_atom(self, atom: Union[Query, Element, int, str], *args,
                 neighbors: Union[int, List[int], Tuple[int, ...], None] = None,
                 hybridization: Union[int, List[int], Tuple[int, ...], None] = None,
                 hydrogens: Union[int, List[int], Tuple[int, ...], None] = None,
                 rings_sizes: Union[int, List[int], Tuple[int, ...], None] = None,
                 heteroatoms: Union[int, List[int], Tuple[int, ...], None] = None,
                 masked: bool = False, **kwargs):
        if hybridization is None:
            hybridization = ()
        elif isinstance(hybridization, int):
            if hybridization < 1 or hybridization > 4:
                raise ValueError('hybridization should be in range [1, 4]')
            hybridization = (hybridization,)
        elif isinstance(hybridization, (tuple, list)):
            if not all(isinstance(h, int) for h in hybridization):
                raise TypeError('hybridizations should be list or tuple of ints')
            if any(h < 1 or h > 4 for h in hybridization):
                raise ValueError('hybridizations should be in range [1, 4]')
            if len(set(hybridization)) != len(hybridization):
                raise ValueError('hybridizations should be unique')
            hybridization = tuple(sorted(hybridization))
        else:
            raise TypeError('hybridization should be int or list or tuple of ints')

        if rings_sizes is None:
            rings_sizes = ()
        elif isinstance(rings_sizes, int):
            if rings_sizes < 3 and rings_sizes != 0:
                raise ValueError('rings should be greater or equal 3. ring equal to zero is no ring atom mark')
            rings_sizes = (rings_sizes,)
        elif isinstance(rings_sizes, (tuple, list)):
            if not all(isinstance(n, int) for n in rings_sizes):
                raise TypeError('rings should be list or tuple of ints')
            if any(n < 3 for n in rings_sizes):
                raise ValueError('rings should be greater or equal 3')
            if len(set(rings_sizes)) != len(rings_sizes):
                raise ValueError('rings should be unique')
            rings_sizes = tuple(sorted(rings_sizes))
        else:
            raise TypeError('rings should be int or list or tuple of ints')

        neighbors = _validate_neighbors(neighbors)
        hydrogens = _validate_neighbors(hydrogens)
        heteroatoms = _validate_neighbors(heteroatoms)

        if not isinstance(atom, Query):
            if isinstance(atom, Element):
                atom = QueryElement.from_atomic_number(atom.atomic_number)(atom.isotope)
            elif isinstance(atom, str):
                atom = QueryElement.from_symbol(atom)()
            elif isinstance(atom, int):
                atom = QueryElement.from_atomic_number(atom)()
            else:
                raise TypeError('QueryElement object expected')

        n = super().add_atom(atom, *args, **kwargs)
        self._neighbors[n] = neighbors
        self._hybridizations[n] = hybridization
        self._hydrogens[n] = hydrogens
        self._rings_sizes[n] = rings_sizes
        self._heteroatoms[n] = heteroatoms
        self._masked[n] = masked
        return n

    def add_bond(self, n, m, bond: Union[QueryBond, Bond, int, Tuple[int, ...]]):
        if isinstance(bond, Bond):
            bond = QueryBond.from_bond(bond)
        elif not isinstance(bond, QueryBond):
            bond = QueryBond(bond)

        sct = self._stereo_cis_trans_paths  # save
        sa = self._stereo_allenes_paths

        super().add_bond(n, m, bond)
        # remove stereo marks on bonded atoms and all its bonds
        if n in self._atoms_stereo:
            del self._atoms_stereo[n]
        if m in self._atoms_stereo:
            del self._atoms_stereo[m]
        if self._cis_trans_stereo:
            for nm, path in sct.items():
                if (n in path or m in path) and nm in self._cis_trans_stereo:
                    del self._cis_trans_stereo[nm]
        if self._allenes_stereo:
            for c, path in sa.items():
                if (n in path or m in path) and c in self._allenes_stereo:
                    del self._allenes_stereo[c]

    def copy(self) -> 'QueryContainer':
        copy = super().copy()

        copy._bonds = cb = {}
        for n, m_bond in self._bonds.items():
            cb[n] = cbn = {}
            for m, bond in m_bond.items():
                if m in cb:  # bond partially exists. need back-connection.
                    cbn[m] = cb[m][n]
                else:
                    cbn[m] = bond.copy()

        copy._neighbors = self._neighbors.copy()
        copy._hybridizations = self._hybridizations.copy()
        copy._hydrogens = self._hydrogens.copy()
        copy._heteroatoms = self._heteroatoms.copy()
        copy._rings_sizes = self._rings_sizes.copy()
        copy._atoms_stereo = self._atoms_stereo.copy()
        copy._allenes_stereo = self._allenes_stereo.copy()
        copy._cis_trans_stereo = self._cis_trans_stereo.copy()
        copy._masked = self._masked.copy()
        return copy

    def union(self, other: 'QueryContainer') -> 'QueryContainer':
        if not isinstance(other, QueryContainer):
            raise TypeError('QueryContainer expected')
        u = super().union(other)

        ub = u._bonds
        for n, m_bond in other._bonds.items():
            ub[n] = ubn = {}
            for m, bond in m_bond.items():
                if m in ub:  # bond partially exists. need back-connection.
                    ubn[m] = ub[m][n]
                else:
                    ubn[m] = bond.copy()

        u._neighbors.update(other._neighbors)
        u._hybridizations.update(other._hybridizations)
        u._hydrogens.update(other._hydrogens)
        u._rings_sizes.update(other._rings_sizes)
        u._atoms_stereo.update(other._atoms_stereo)
        u._allenes_stereo.update(other._allenes_stereo)
        u._cis_trans_stereo.update(other._cis_trans_stereo)
        u._heteroatoms.update(other._heteroatoms)
        u._masked.update(other._masked)
        return u

    def remap(self, mapping: Dict[int, int], *, copy=False) -> 'QueryContainer':
        """
        Change atom numbers.

        :param copy: keep original graph.
        """
        if len(mapping) != len(set(mapping.values())) or \
                not (self._atoms.keys() - mapping.keys()).isdisjoint(mapping.values()):
            raise ValueError('mapping overlap')

        mg = mapping.get
        charges = self._charges
        radicals = self._radicals
        hydrogens = self._hydrogens
        neighbors = self._neighbors
        hybridizations = self._hybridizations
        heteroatoms = self._heteroatoms
        rings_sizes = self._rings_sizes
        masked = self._masked

        if copy:
            h = self.__class__()
            ha = h._atoms
            hb = h._bonds
            hc = h._charges
            hr = h._radicals
            hhg = h._hydrogens
            hn = h._neighbors
            hh = h._hybridizations
            hx = h._heteroatoms
            hrs = h._rings_sizes
            hm = h._masked
            has = h._atoms_stereo
            hal = h._allenes_stereo
            hcs = h._cis_trans_stereo

            for n, atom in self._atoms.items():
                m = mg(n, n)
                atom = atom.copy()
                ha[m] = atom
                atom._attach_graph(h, m)

            # deep copy of bonds
            for n, m_bond in self._bonds.items():
                n = mg(n, n)
                hb[n] = hbn = {}
                for m, bond in m_bond.items():
                    m = mg(m, m)
                    if m in hb:  # bond partially exists. need back-connection.
                        hbn[m] = hb[m][n]
                    else:
                        hbn[m] = bond.copy()
        else:
            ha = {}
            hb = {}
            hc = {}
            hr = {}
            hhg = {}
            hn = {}
            hh = {}
            hx = {}
            hrs = {}
            hm = {}
            has = {}
            hal = {}
            hcs = {}

            for n, atom in self._atoms.items():
                m = mg(n, n)
                ha[m] = atom
                atom._change_map(m)  # change mapping number

            for n, m_bond in self._bonds.items():
                n = mg(n, n)
                hb[n] = hbn = {}
                for m, bond in m_bond.items():
                    m = mg(m, m)
                    if m in hb:  # bond partially exists. need back-connection.
                        hbn[m] = hb[m][n]
                    else:
                        hbn[m] = bond

        for n in self._atoms:
            m = mg(n, n)
            hc[m] = charges[n]
            hr[m] = radicals[n]
            hhg[m] = hydrogens[n]
            hn[m] = neighbors[n]
            hh[m] = hybridizations[n]
            hx[m] = heteroatoms[n]
            hrs[m] = rings_sizes[n]
            hm[m] = masked[n]

        for n, stereo in self._atoms_stereo.items():
            has[mg(n, n)] = stereo
        for n, stereo in self._allenes_stereo.items():
            hal[mg(n, n)] = stereo
        for (n, m), stereo in self._cis_trans_stereo.items():
            hcs[(mg(n, n), mg(m, m))] = stereo

        if copy:
            return h

        self._atoms = ha
        self._bonds = hb
        self._charges = hc
        self._radicals = hr
        self._hydrogens = hhg
        self._neighbors = hn
        self._hybridizations = hh
        self._heteroatoms = hx
        self._rings_sizes = hrs
        self._masked = hm
        self._atoms_stereo = has
        self._allenes_stereo = hal
        self._cis_trans_stereo = hcs
        self.flush_cache()
        return self

    def get_mapping(self, other: Union['QueryContainer', 'molecule.MoleculeContainer'], /, *, _cython=True, **kwargs):
        # _cython - by default cython implementation enabled.
        # disable it by overriding method if Query Atoms or Containers logic changed.
        # Lv, Ts and Og in cython optimized mode treated as equal.
        if isinstance(other, QueryContainer):
            return super().get_mapping(other, **kwargs)
        elif isinstance(other, molecule.MoleculeContainer):
            return super().get_mapping(other, _cython=_cython, **kwargs)
        raise TypeError('MoleculeContainer or QueryContainer expected')

    def enumerate_queries(self, *, enumerate_marks: bool = False):
        """
        Enumerate complex queries into multiple simple ones. For example `[N,O]-C` into `NC` and `OC`.

        :param enumerate_marks: enumerate multiple marks to separate queries
        """
        atoms = [(n, a._numbers) for n, a in self._atoms.items() if isinstance(a, ListElement)]
        bonds = [(n, m, b.order) for n, m, b in self.bonds() if len(b.order) > 1]
        for combo in product(*(x for *_, x in chain(atoms, bonds))):
            copy = self.copy()
            for (n, _), a in zip(atoms, combo):
                copy._atoms[n] = a = QueryElement.from_atomic_number(a)()
                a._attach_graph(copy, n)
            for (n, m, _), b in zip(bonds, combo[len(atoms):]):
                copy._bonds[n][m]._QueryBond__order = (b,)

            if enumerate_marks:
                c = 0
                slices = []
                data = []
                for attr in ('_neighbors', '_hybridizations', '_hydrogens', '_heteroatoms', '_rings_sizes'):
                    tmp = [(n, v) for n, v in getattr(self, attr).items() if len(v) > 1]
                    if tmp:
                        data.extend(tmp)
                        slices.append((attr, c, c + len(tmp)))
                        c += len(tmp)

                for combo2 in product(*(x for _, x in data)):
                    copy2 = copy.copy()
                    for attr, i, j in slices:
                        attr = getattr(copy2, attr)
                        for (n, _), v in zip(data[i: j], combo2[i: j]):
                            attr[n] = (v,)
                    yield copy2
            else:
                yield copy

    @cached_property
    def _cython_compiled_query(self):
        # long I:
        # bond: single, double, triple, aromatic, special = 5 bit
        # bond in ring: 2 bit
        # atom: H-Ba: 56 bit
        # transfer bit

        # long II:
        # atom La-Mc: 59 bit
        # Lv-Ts-Og: 3 elements packed into 1 bit.
        # hybridizations: 1-4 = 4 bit

        # long III:
        # isotope: not specified, isotope - common_isotope = -8 - +8 = 18 bit
        # is_radical: 2 bit
        # charge: -4 - +4: 9 bit
        # implicit_hydrogens: 0-4 = 5 bit
        # neighbors: 0-14 = 15 bit
        # heteroatoms: 0-14 = 15 bit

        # long IV:
        # ring_sizes: not-in-ring bit, 3-atom ring, 4-...., 65-atom ring

        # int V: bonds closures
        # padding: 1 bit
        # bond: single, double, triple, aromatic, special = 5 bit
        # bond in ring: 2 bit
        from ..files._mdl.mol import common_isotopes

        _components, _closures = self._compiled_query
        components = []
        for c in _components:
            mapping = {n: i for i, (n, *_) in enumerate(c)}
            masks1 = []
            masks2 = []
            masks3 = []
            masks4 = []
            for *_, a, b in c:
                if isinstance(a, AnyMetal):  # isotope, radical, charge, hydrogens and heteroatoms states ignored
                    # elements except 1, 2, 5, 6, 7, 8, 9, 10, 14, 15, 16, 17, 18, 32, 33, 34, 35, 36, 51, 52, 53, 54
                    v1 = 0x0060707ffc1fff87
                    v2 = 0xfffffffffffffff0
                    v3 = 0xffffffffc0007fff
                    v4 = 0xffffffffffffffff
                else:
                    if isinstance(a, AnyElement):
                        v1 = 0x01ffffffffffffff
                        v2 = 0xfffffffffffffff0
                    else:
                        if isinstance(a, ListElement):
                            v1 = v2 = 0
                            for n in a._numbers:
                                if n > 56:
                                    if n > 116:  # Ts, Og
                                        n = 116
                                    v1 |= 1  # set transfer bit
                                    v2 |= 1 << (120 - n)
                                else:
                                    v1 |= 1 << (57 - n)
                        elif (n := a.atomic_number) > 56:
                            if n > 116:  # Ts, Og
                                n = 116
                            v1 = 1  # transfer bit
                            v2 = 1 << (120 - n)
                        else:
                            v1 = 1 << (57 - n)
                            v2 = 0
                    if a.isotope:
                        v3 = 1 << (a.isotope - common_isotopes[a.atomic_symbol] + 54)
                        if a.is_radical:
                            v3 |= 0x200000000000
                        else:
                            v3 |= 0x100000000000
                    elif a.is_radical:  # any isotope
                        v3 = 0xffffe00000000000
                    else:
                        v3 = 0xffffd00000000000

                    v3 |= 1 << (a.charge + 39)

                    if not a.implicit_hydrogens:
                        v3 |= 0x7c0000000
                    else:
                        for h in a.implicit_hydrogens:
                            v3 |= 1 << (h + 30)

                    if not a.heteroatoms:
                        v3 |= 0x7fff
                    else:
                        for n in a.heteroatoms:
                            v3 |= 1 << n

                    if a.ring_sizes:
                        if a.ring_sizes[0]:
                            v4 = 0
                            for r in a.ring_sizes:
                                if r > 65:  # big rings not supported
                                    continue
                                v4 |= 1 << (65 - r)
                            if not v4:  # only 65+ rings. set as rings-free.
                                v4 = 0x8000000000000000
                        else:  # not in rings
                            v4 = 0x8000000000000000
                    else:  # any rings
                        v4 = 0xffffffffffffffff

                if not a.neighbors:
                    v3 |= 0x3fff8000
                else:
                    for n in a.neighbors:
                        v3 |= 1 << (n + 15)

                if not a.hybridization:
                    v2 |= 0xf
                else:
                    for n in a.hybridization:
                        v2 |= 1 << (n - 1)

                if b is not None:
                    for o in b.order:
                        if o == 1:
                            v1 |= 0x0800000000000000
                        elif o == 4:
                            v1 |= 0x4000000000000000
                        elif o == 2:
                            v1 |= 0x1000000000000000
                        elif o == 3:
                            v1 |= 0x2000000000000000
                        else:
                            v1 |= 0x8000000000000000
                    if b.in_ring is None:
                        v1 |= 0x0600000000000000
                    elif b.in_ring:
                        v1 |= 0x0400000000000000
                    else:
                        v1 |= 0x0200000000000000

                masks1.append(v1)
                masks2.append(v2)
                masks3.append(v3)
                masks4.append(v4)

            closures = [0] * len(c)  # closures amount
            q_from = [0] * len(c)
            q_to = [0] * len(c)
            indices = [0] * sum(len(ms) for n, ms in _closures.items() if n in mapping)
            bonds = indices.copy()

            start = 0
            for n, ms in _closures.items():
                if (i := mapping.get(n)) is not None:
                    closures[i] = len(ms)
                    q_from[i] = start
                    for j, (m, b) in enumerate(ms, start):
                        v = 0x01ffffffffffffff  # atom doesn't matter.
                        for o in b.order:
                            if o == 1:
                                v |= 0x0800000000000000
                            elif o == 4:
                                v |= 0x4000000000000000
                            elif o == 2:
                                v |= 0x1000000000000000
                            elif o == 3:
                                v |= 0x2000000000000000
                            else:
                                v |= 0x8000000000000000
                        if b.in_ring is None:
                            v |= 0x0600000000000000
                        elif b.in_ring:
                            v |= 0x0400000000000000
                        else:
                            v |= 0x0200000000000000
                        bonds[j] = v
                        indices[j] = mapping[m]
                    start += len(ms)
                    q_to[i] = start
            components.append((array('L', [n for n, *_ in c]), array('I', [0] + [mapping[x] for _, x, *_ in c[1:]]),
                               array('Q', masks1), array('Q', masks2), array('Q', masks3), array('Q', masks4),
                               array('I', closures), array('I', q_from), array('I', q_to),
                               array('I', indices), array('Q', bonds)))
        return components

    def __getstate__(self):
        return {'atoms_stereo': self._atoms_stereo, 'allenes_stereo': self._allenes_stereo,
                'cis_trans_stereo': self._cis_trans_stereo, 'neighbors': self._neighbors,
                'hybridizations': self._hybridizations, 'hydrogens': self._hydrogens, 'masked': self._masked,
                'rings_sizes': self._rings_sizes, 'heteroatoms': self._heteroatoms, **super().__getstate__()}

    def __setstate__(self, state):
        super().__setstate__(state)
        self._atoms_stereo = state['atoms_stereo']
        self._allenes_stereo = state['allenes_stereo']
        self._cis_trans_stereo = state['cis_trans_stereo']
        self._neighbors = state['neighbors']
        self._hybridizations = state['hybridizations']
        self._hydrogens = state['hydrogens']
        self._rings_sizes = state['rings_sizes']
        self._heteroatoms = state['heteroatoms']
        self._masked = state['masked']


__all__ = ['QueryContainer']



# -*- coding: utf-8 -*-
#
#  Copyright 2019-2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from importlib.util import find_spec
from .free_wilson import *
from .functional_groups import *
from .grid import *
from .retro import *
from .svg import *
from ..containers.graph import Graph


_patched = False


def patch_pandas():
    """
    Fix pandas molecules representation.
    """
    # prevent recursive patching
    global _patched
    if _patched:
        return
    _patched = True

    from pandas.io.formats import printing
    from pandas.io.formats.printing import is_sequence

    def w(obj):
        if isinstance(obj, Graph):
            return False
        return is_sequence(obj)

    printing.is_sequence = w


__all__ = ['functional_groups', 'fw_prepare_groups', 'fw_decomposition_tree',
           'grid_depict', 'GridDepict', 'retro_depict', 'RetroDepict', 'svg2png', 'patch_pandas']


if find_spec('rdkit'):
    from .rdkit import *
    __all__.extend(['from_rdkit_molecule', 'to_rdkit_molecule'])



# -*- coding: utf-8 -*-
#
#  Copyright 2020-2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2020 Nail Samikaev <samikaevn@yandex.ru>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from collections import deque
from typing import TYPE_CHECKING, Iterator, Union
from .acid_base import *
from .heteroarenes import *
from .keto_enol import *


if TYPE_CHECKING:
    from chython import MoleculeContainer


class Tautomers(AcidBase, HeteroArenes, KetoEnol):
    """
    Oxides and sulphides ignored.
    """
    __slots__ = ()

    def enumerate_tautomers(self: Union['MoleculeContainer', 'Tautomers'], *, prepare_molecules=True, zwitter=True,
                            partial=False, increase_aromaticity=True, keep_sugars=True, heteroarenes=True,
                            keto_enol=True, limit: int = 1000) -> Iterator['MoleculeContainer']:
        """
        Enumerate all possible tautomeric forms of molecule.

        :param prepare_molecules: Standardize structures for correct processing
        :param zwitter: Do zwitter-ions enumeration
        :param partial: Allow OC=CC=C>>O=CCC=C or O=CC=CC>>OC=C=CC
        :param increase_aromaticity: prevent aromatic ring destruction
        :param keep_sugars: prevent carbonyl moving in sugars
        :param heteroarenes: enumerate heteroarenes
        :param keto_enol: enumerate keto-enols
        :param limit: Maximum attempts count
        """
        if limit < 1:
            raise ValueError('limit should be greater or equal 1')

        has_stereo = bool(self._atoms_stereo or self._allenes_stereo or self._cis_trans_stereo)
        counter = 0

        copy = self.copy()
        copy.clean_stereo()
        # sssr, neighbors and heteroatoms are same for all tautomers.
        # prevent recalculation by sharing cache.
        self.__set_cache(copy)
        if prepare_molecules:  # transform to kekule form without hydrogens
            k = copy.kekule()
            i = copy.implicify_hydrogens(_fix_stereo=False)
            if k or i:  # reset cache after flush
                self.__set_cache(copy)

        thiele = copy.copy()  # transform to thiele to prevent duplicates and dearomatization
        self.__set_cache(thiele)
        if thiele.thiele(fix_tautomers=False):
            self.__set_cache(thiele)

        # return origin structure as first tautomer
        if has_stereo:
            yield self.__set_stereo(thiele.copy())
        else:
            yield thiele

        seen = {thiele: None}  # value is parent molecule - required for preventing migrations in sugars.

        # first try to neutralize
        if copy.neutralize(_fix_stereo=False):  # found neutral form
            thiele = copy.copy()
            self.__set_cache(copy)  # restore cache
            self.__set_cache(thiele)
            if thiele.thiele(fix_tautomers=False):
                self.__set_cache(thiele)

            # return found neutral form
            if has_stereo:
                yield self.__set_stereo(thiele.copy())
            else:
                yield thiele
            counter += 1
            seen[thiele] = None

        # lets iteratively do keto-enol transformations.
        rings_count = len(thiele.aromatic_rings)  # increase rings strategy.
        if keto_enol:
            queue = deque([(copy, thiele)])
        else:
            queue = None
        new_queue = [thiele]  # new_queue - molecules suitable for hetero-arenes enumeration.
        # store aromatic form to seen. kekule forms not suitable for duplicate checking.

        while queue:
            current, thiele_current = queue.popleft()
            for mol, ket in current._enumerate_keto_enol_tautomers(partial):
                thiele = mol.copy()
                self.__set_cache(mol)
                self.__set_cache(thiele)
                if thiele.thiele(fix_tautomers=False):  # reset cache after flush_cache.
                    self.__set_cache(thiele)

                if thiele not in seen:
                    seen[thiele] = current
                    rc = len(thiele.aromatic_rings)
                    if increase_aromaticity:
                        if rc < rings_count:  # skip aromatic rings destruction
                            continue
                        elif rc > rings_count:  # higher aromaticity found. flush old queues.
                            rings_count = rc
                            queue = deque([(mol, thiele)])
                            new_queue = [thiele]
                            copy = mol  # new entry point.
                            if has_stereo:
                                yield self.__set_stereo(thiele.copy())
                            else:
                                yield thiele
                            break
                    if keep_sugars and current is not copy and ket:
                        # prevent carbonyl migration in sugars. skip entry point.
                        # search alpha hydroxy ketone inversion
                        before = seen[thiele_current]._sugar_groups
                        if any((k, e) in before for e, k in mol._sugar_groups):
                            continue

                    queue.append((mol, thiele))
                    new_queue.append(thiele)
                    if has_stereo:
                        yield self.__set_stereo(thiele.copy())
                    else:
                        yield thiele
                counter += 1
                if counter == limit:
                    return

        # heteroarenes tautomers enumeration
        if heteroarenes:
            queue = deque(new_queue)
        while queue:
            current = queue.popleft()
            for mol in current._enumerate_hetero_arene_tautomers():
                self.__set_cache(mol)
                if mol not in seen:
                    seen[mol] = None
                    queue.append(mol)
                    new_queue.append(mol)  # new hetero-arenes also should be included to this list.
                    if has_stereo:
                        yield self.__set_stereo(mol.copy())
                    else:
                        yield mol
                counter += 1
                if counter == limit:
                    return

        # zwitter-ions enumeration
        if zwitter:
            queue = deque(new_queue)
        while queue:
            current = queue.popleft()
            for mol in current._enumerate_zwitter_tautomers():
                self.__set_cache(mol)
                if mol not in seen:
                    seen[mol] = None
                    queue.append(mol)
                    if has_stereo:
                        yield self.__set_stereo(mol.copy())
                    else:
                        yield mol
                counter += 1
                if counter == limit:
                    return

    def enumerate_charged_tautomers(self: 'MoleculeContainer', *, prepare_molecules=True, partial=False,
                                    increase_aromaticity=True, keep_sugars=True, heteroarenes=True,
                                    keto_enol=True, deep: int = 4, limit: int = 1000):
        """
        Enumerate tautomers and protonated-deprotonated forms.
        Better to use on neutralized non-ionic molecules.

        See `enumerate_tautomers` and `enumerate_charged_forms` params description.
        """
        count = 0
        for t in self.enumerate_tautomers(prepare_molecules=prepare_molecules, zwitter=False, partial=partial,
                                          increase_aromaticity=increase_aromaticity, keep_sugars=keep_sugars,
                                          heteroarenes=heteroarenes, keto_enol=keto_enol, limit=limit):
            yield t
            count += 1
            if count == limit:
                return
            for c in t.enumerate_charged_forms(deep=deep, limit=limit):
                yield c
                count += 1
                if count == limit:
                    return

    def __set_cache(self: 'MoleculeContainer', mol):
        try:
            neighbors = self.__dict__['__cached_args_method_neighbors']
        except KeyError:
            neighbors = self.__dict__['__cached_args_method_neighbors'] = {}
        try:
            heteroatoms = self.__dict__['__cached_args_method_heteroatoms']
        except KeyError:
            heteroatoms = self.__dict__['__cached_args_method_heteroatoms'] = {}
        try:
            is_ring_bond = self.__dict__['__cached_args_method_is_ring_bond']
        except KeyError:
            is_ring_bond = self.__dict__['__cached_args_method_is_ring_bond'] = {}

        mol.__dict__['sssr'] = self.sssr  # thiele/kekule
        mol.__dict__['ring_atoms'] = self.ring_atoms  # morgan
        mol.__dict__['_connected_components'] = self._connected_components  # isomorphism
        mol.__dict__['atoms_rings_sizes'] = self.atoms_rings_sizes  # isomorphism
        mol.__dict__['__cached_args_method_neighbors'] = neighbors  # isomorphism
        mol.__dict__['__cached_args_method_heteroatoms'] = heteroatoms  # isomorphism
        mol.__dict__['__cached_args_method_is_ring_bond'] = is_ring_bond  # isomorphism

    def __set_stereo(self: 'MoleculeContainer', mol):
        mol._atoms_stereo.update(self._atoms_stereo)
        mol._allenes_stereo.update(self._allenes_stereo)
        mol._cis_trans_stereo.update(self._cis_trans_stereo)
        mol.fix_stereo()
        return mol


__all__ = ['Tautomers']



# -*- coding: utf-8 -*-
#
#  Copyright 2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from collections import defaultdict
from ...exceptions import IncorrectSmiles


def parser(tokens, strong_cycle):
    log = []
    t1 = tokens[0][0]
    if t1 == 2:
        if tokens[1][0] not in (0, 8):
            raise IncorrectSmiles('not atom started')
    elif t1 not in (0, 8):
        raise IncorrectSmiles('not atom started')

    atoms = []
    bonds = []
    order = defaultdict(list)
    atoms_types = []
    atom_num = 0
    last_num = 0
    stack = []
    cycles = {}
    used_cycles = set()
    stereo_bonds = defaultdict(dict)
    previous = None

    for token_type, token in tokens:
        if token_type == 2:  # ((((((
            if previous:
                if previous[0] != 4:
                    raise IncorrectSmiles('bond before side chain')
                previous = None
            stack.append(last_num)
        elif token_type == 3:  # ))))))
            if previous:
                raise IncorrectSmiles('bond before closure')
            try:
                last_num = stack.pop()
            except IndexError:
                raise IncorrectSmiles('close chain more than open')
        elif token_type in (1, 4, 9, 10, 12):  # bonds. only keeping for atoms connecting
            if previous:
                raise IncorrectSmiles('2 bonds in a row')
            elif not atoms:
                raise IncorrectSmiles('started from bond')
            previous = (token_type, token)
        elif token_type == 6:  # cycle
            if previous and previous[0] == 4:
                raise IncorrectSmiles('dot-cycle pattern invalid')
            elif token not in cycles:
                if token in used_cycles:
                    if strong_cycle:
                        raise IncorrectSmiles('reused closure number')
                    else:
                        log.append(f'reused closure number: {token}')
                else:
                    used_cycles.add(token)
                cycles[token] = (last_num, previous, len(order[last_num]))
                order[last_num].append(None)  # Reserve a table
                previous = None
            else:
                a, ob, ind = cycles[token]
                if ob:
                    if not previous:
                        bt, b = ob
                        if bt == 9:  # closure open is \/ bonded
                            stereo_bonds[a][last_num] = b
                            b = 1
                        elif strong_cycle:
                            raise IncorrectSmiles('not equal cycle bonds')
                        else:
                            log.append('ignored difference in cycle bonds')
                    else:
                        bt, b = previous
                        obt, ob = ob
                        if bt == 9:  # \/ bonds can be unequal
                            if obt == 9:
                                stereo_bonds[a][last_num] = ob
                            elif ob != 1:
                                raise IncorrectSmiles('not equal cycle bonds')
                            stereo_bonds[last_num][a] = b
                            b = 1
                        elif obt == 9:
                            if b != 1:
                                raise IncorrectSmiles('not equal cycle bonds')
                            stereo_bonds[a][last_num] = ob
                        elif b != ob:
                            raise IncorrectSmiles('not equal cycle bonds')
                        previous = None
                elif previous:
                    bt, b = previous
                    if bt == 9:  # stereo \/
                        stereo_bonds[last_num][a] = b
                        b = 1
                    elif strong_cycle:
                        raise IncorrectSmiles('not equal cycle bonds')
                    else:
                        log.append('ignored difference in cycle bonds')
                    previous = None
                else:
                    b = 4 if atoms_types[last_num] == atoms_types[a] == 8 else 1

                bonds.append((last_num, a, b))
                order[a][ind] = last_num
                order[last_num].append(a)
                del cycles[token]
        else:  # atom
            if atoms:
                if not previous:
                    bonds.append((atom_num, last_num, 4 if token_type == atoms_types[last_num] == 8 else 1))
                    order[last_num].append(atom_num)
                    order[atom_num].append(last_num)
                else:
                    bt, b = previous
                    if bt == 9:
                        bonds.append((atom_num, last_num, 4 if token_type == atoms_types[last_num] == 8 else 1))
                        order[last_num].append(atom_num)
                        order[atom_num].append(last_num)

                        stereo_bonds[last_num][atom_num] = b
                        stereo_bonds[atom_num][last_num] = not b
                    elif bt in (1, 10, 12):
                        bonds.append((atom_num, last_num, b))
                        order[last_num].append(atom_num)
                        order[atom_num].append(last_num)
                    # else bt == 4 - skip dot
                    previous = None

            atoms.append(token)
            atoms_types.append(token_type)
            last_num = atom_num
            atom_num += 1

    if stack:
        raise IncorrectSmiles('number of ( does not equal to number of )')
    elif cycles:
        raise IncorrectSmiles('cycle is not finished')
    elif previous:
        raise IncorrectSmiles('bond on the end')

    stereo_bonds = {n: ms for n, ms in stereo_bonds.items() if len(ms) == 1 or len(ms) == set(ms.values())}
    mol = {'atoms': atoms, 'bonds': bonds, 'order': order, 'stereo_bonds': stereo_bonds, 'meta': {}}
    return mol, log


__all__ = ['parser']



# -*- coding: utf-8 -*-
#
#  Copyright 2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from collections import ChainMap
from itertools import count
from typing import List, Tuple, TYPE_CHECKING, Union
from ._reactions import rules


if TYPE_CHECKING:
    from chython import ReactionContainer


class FixMapper:
    __slots__ = ()

    def fix_mapping(self: 'ReactionContainer', *, logging: bool = False) -> \
            Union[bool, List[Tuple[int, str, Tuple[int, ...]]]]:
        """
        Fix mapping by using loaded rules.
        """
        if not self:
            if logging:
                return []
            return False

        cgr = ~self
        if not cgr.center_atoms:
            if logging:
                return []
            return False
        del self.__dict__['__cached_method_compose']

        log = []
        free_number = count(max(cgr) + 1)
        components = [(cgr.substructure(c),
                       cgr.augmented_substructure(c, 2),  # deep DEPENDS on rules!
                       set(c))
                      for c in cgr.substructure(cgr.center_atoms).connected_components]

        r_atoms = ChainMap(*(x._atoms for x in self.reactants))
        for c, ac, cs in components:
            for rule_num, (query, signature, restrict, fix) in enumerate(rules):
                if str(c) == signature:
                    for mapping in query.get_mapping(ac, automorphism_filter=False):
                        if not cs.issubset(mapping.values()):
                            continue
                        if restrict is not None and any(a != r_atoms.get(mapping[n]) for n, a in restrict.atoms()):
                            continue
                        mapping = {mapping[n]: next(free_number) if m is None else mapping[m] for n, m in fix.items()}
                        for m in self.products:
                            m.remap(mapping)
                        log.append((rule_num, signature, tuple(mapping.values())))
                        break
                    else:
                        continue
                    break  # component remapped!

        if log:
            self.flush_cache()
            if logging:
                return log
            return True
        elif logging:
            return log
        return False


__all__ = ['FixMapper']



# -*- coding: utf-8 -*-
#
#  Copyright 2021, 2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2021 Aleksandr Sizov <murkyrussian@gmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
cimport cython
from cpython.mem cimport PyMem_Malloc, PyMem_Free
from libc.string cimport memset

cdef extern from "Python.h":
    dict _PyDict_NewPresized(Py_ssize_t minused)


@cython.boundscheck(False)
@cython.wraparound(False)
def get_mapping(unsigned long[::1] q_numbers not None, unsigned int[::1] q_back not None,
                unsigned long long[::1] q_masks1 not None, unsigned long long[::1] q_masks2 not None,
                unsigned long long[::1] q_masks3 not None, unsigned long long[::1] q_masks4 not None,
                unsigned int[::1] q_closures not None, unsigned int[::1] q_from not None,
                unsigned int[::1] q_to not None, unsigned int[::1] q_indices not None,
                unsigned long long[::1] q_bonds not None, unsigned long[::1] o_numbers not None,
                unsigned long long[::1] o_bits1 not None, unsigned long long[::1] o_bits2 not None,
                unsigned long long[::1] o_bits3 not None, unsigned long long[::1] o_bits4 not None,
                unsigned long long[::1] o_bonds not None, unsigned int[::1] o_from not None,
                unsigned int[::1] o_to not None, unsigned int[::1] o_indices not None,
                unsigned int[::1] scope not None):
    # expected less than 2^16 atoms in structure.
    cdef unsigned int stack = 0, path_size = 0, q_size, q_size_dec, o_size, depth, front, back, closures_num
    cdef unsigned int n, m, o, i, j, closures_counter
    cdef unsigned long long q_mask1, q_mask2, q_mask3, q_mask4, o_bond, c_bond
    cdef dict mapping

    q_size = len(q_numbers)
    q_size_dec = q_size - 1
    o_size = len(o_numbers)
    cdef unsigned int *path = <unsigned int *> PyMem_Malloc(q_size_dec * sizeof(unsigned int))
    cdef unsigned int *stack_index = <unsigned int *> PyMem_Malloc(2 * o_size * sizeof(unsigned int))
    cdef unsigned int *stack_depth = <unsigned int *> PyMem_Malloc(2 * o_size * sizeof(unsigned int))
    cdef bint *matched = <bint *> PyMem_Malloc(o_size * sizeof(bint))
    cdef unsigned long long *o_closures = <unsigned long long *> PyMem_Malloc(o_size * sizeof(unsigned long long))

    if not path or not stack_index or not stack_depth or not matched or not o_closures:
        raise MemoryError()

    memset(matched, 0, o_size * sizeof(bint))
    memset(o_closures, 0, o_size * sizeof(unsigned long long))

    # find entry-points.
    q_mask1 = q_masks1[0]
    q_mask2 = q_masks2[0]
    q_mask3 = q_masks3[0]
    q_mask4 = q_masks4[0]
    for n in range(o_size):
        if (scope[n] and
            q_mask1 & o_bits1[n] and  # o_bits1 doesn't contain bond bits.
            q_mask2 & o_bits2[n] == o_bits2[n] and
            q_mask3 & o_bits3[n] == o_bits3[n] and
            q_mask4 & o_bits4[n]):

            stack_index[stack] = n
            stack_depth[stack] = 0
            stack += 1

    try:
        while stack:
            stack -= 1
            depth = stack_depth[stack]
            n = stack_index[stack]

            if depth == q_size_dec:
                mapping = _PyDict_NewPresized(q_size)
                for i in range(depth):
                    mapping[q_numbers[i]] = o_numbers[path[i]]
                mapping[q_numbers[depth]] = o_numbers[n]
                yield mapping
            else:
                if path_size != depth:  # dead end reached
                    for i in range(depth, path_size):
                        matched[path[i]] = False  # mark unmatched
                    path_size = depth

                matched[n] = True
                path[path_size] = n
                path_size += 1

                front = depth + 1
                back = q_back[front]
                if back != depth:  # branch
                    n = path[back]

                # load next query atom
                q_mask1 = q_masks1[front]
                q_mask2 = q_masks2[front]
                q_mask3 = q_masks3[front]
                q_mask4 = q_masks4[front]
                closures_num = q_closures[front]

                for i in range(o_from[n], o_to[n]):
                    o_bond = o_bonds[i]
                    m = o_indices[i]
                    if (scope[m] and not matched[m] and
                        q_mask1 & o_bond == o_bond and  # bond order, in ring mark and atom bit should match.
                        q_mask2 & o_bits2[m] == o_bits2[m] and
                        q_mask3 & o_bits3[m] == o_bits3[m] and
                        q_mask4 & o_bits4[m]):

                        if closures_num:  # candidate atom should have same closures.
                            closures_counter = 0
                            # make a map of closures for o_n atom
                            # an index is a neighbor atom and a value is a bond between o_n and the neighbor
                            for j in range(o_from[m], o_to[m]):
                                o = o_indices[j]
                                if o != n and matched[o]:
                                    o_closures[o] = o_bonds[j]
                                    closures_counter += 1

                            if closures_counter == closures_num:
                                for j in range(q_from[front], q_to[front]):
                                    c_bond = o_closures[path[q_indices[j]]]
                                    if not c_bond or q_bonds[j] & c_bond != c_bond:  # compare order and ring bits
                                        break
                                else:
                                    stack_index[stack] = m
                                    stack_depth[stack] = front
                                    stack += 1

                            # fill an array with nulls
                            for j in range(o_from[m], o_to[m]):
                                o_closures[o_indices[j]] = 0
                        else:  # candidate atom should not have closures.
                            for j in range(o_from[m], o_to[m]):
                               o = o_indices[j]
                               if o != n and matched[o]:
                                   break  # found closure
                            else:
                                stack_index[stack] = m
                                stack_depth[stack] = front
                                stack += 1
    finally:
        PyMem_Free(path)
        PyMem_Free(matched)
        PyMem_Free(stack_index)
        PyMem_Free(stack_depth)
        PyMem_Free(o_closures)



# -*- coding: utf-8 -*-
#
#  Copyright 2021, 2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2021 Aleksandr Sizov <murkyrussian@gmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from math import log2
from numpy import zeros, uint8
from typing import Set, TYPE_CHECKING


if TYPE_CHECKING:
    from chython import MoleculeContainer


class MorganFingerprint:
    __slots__ = ()

    def morgan_fingerprint(self, min_radius: int = 1, max_radius: int = 4,
                           length: int = 1024, number_active_bits: int = 2):
        """
        Transform structures into array of binary features.
        Morgan fingerprints. Similar to RDkit implementation.

        :param min_radius: minimal radius of EC
        :param max_radius: maximum radius of EC
        :param length: bit string's length. Should be power of 2
        :param number_active_bits: number of active bits for each hashed tuple

        :return: array(n_features)
        """
        bits = self.morgan_bit_set(min_radius, max_radius, length, number_active_bits)
        fingerprints = zeros(length, dtype=uint8)
        fingerprints[list(bits)] = 1
        return fingerprints

    def morgan_bit_set(self, min_radius: int = 1, max_radius: int = 4,
                       length: int = 1024, number_active_bits: int = 2) -> Set[int]:
        """
        Transform structures into set of indexes of True-valued features.

        :param min_radius: minimal radius of EC
        :param max_radius: maximum radius of EC
        :param length: bit string's length. Should be power of 2
        :param number_active_bits: number of active bits for each hashed tuple
        """
        mask = length - 1
        log = int(log2(length))

        active_bits = set()
        for tpl in self.morgan_hash_set(min_radius, max_radius):
            active_bits.add(tpl & mask)
            if number_active_bits == 2:
                active_bits.add(tpl >> log & mask)
            elif number_active_bits > 2:
                for _ in range(1, number_active_bits):
                    tpl >>= log
                    active_bits.add(tpl & mask)
        return active_bits

    def morgan_hash_set(self: 'MoleculeContainer', min_radius: int = 1, max_radius: int = 4) -> Set[int]:
        """
        Transform structures into integer hashes of atoms with EC.

        :param min_radius: minimal radius of EC
        :param max_radius: maximum radius of EC
        """
        identifiers = {idx: hash((atom.isotope or 0, atom.atomic_number, atom.charge, atom.is_radical))
                       for idx, atom in self.atoms()}

        bonds = self._bonds
        arr = set()
        for step in range(1, max_radius + 1):
            if step >= min_radius:
                arr.update(identifiers.values())
            identifiers = {idx: hash((tpl, *(x for x in
                                             sorted((int(b), identifiers[ngb]) for ngb, b in bonds[idx].items())
                                             for x in x)))
                           for idx, tpl in identifiers.items()}

        if max_radius > 1:  # add last ring
            arr.update(identifiers.values())
        return arr


__all__ = ['MorganFingerprint']



# -*- coding: utf-8 -*-
#
#  Copyright 2021, 2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from collections import defaultdict
from itertools import product
from operator import itemgetter
from random import shuffle
from typing import TYPE_CHECKING, Dict, Optional, Union, List
from ...containers.bonds import Bond
from ...exceptions import ValenceError


if TYPE_CHECKING:
    from chython import MoleculeContainer

# atom, charge, unsaturation
tuned_priority = {(7, 0, 0): -3,  # amine
                  (7, 0, 1): -3,  # X=N-X
                  (7, 0, 2): -3,  # X#N
                  (7, 1, 1): -2,  # X=[N+](-X)-X
                  (7, 1, 2): -2,  # X=[N+]=X
                  (7, -1, 0): -1,  # X-[N-]-X
                  (7, -1, 1): -1,  # X=[N-]
                  (8, 0, 0): -2,  # X-O-X
                  (8, 0, 1): -2,  # X=O
                  (8, -1, 0): -1,  # X-[O-]
                  (8, 1, 1): -1,  # X=[O+]-X
                  (16, 0, 0): -3,  # X-S-X
                  (16, 0, 2): -2,  # X-S(-X)(=X)=X
                  (16, 0, 1): -2,  # X-S(-X)(=X)
                  (16, 0, 2): -2,  # X=S=X
                  (16, 0, 0): -1,  # X-S(-X)(-X)-X
                  (16, 0, 1): -1,  # X-[S+](-X)(-X)
                  (16, 1, 1): -1,  # X=[S+]-X
                  }
charge_priority = {0: 0, -1: 1, 1: 2, 2: 3, 3: 4, -2: 5, 4: 6, -3: 7, -4: 8}


class Saturation:
    __slots__ = ()

    def saturate(self: 'MoleculeContainer', neighbors_distances: Optional[Dict[int, Dict[int, float]]] = None,
                 reset_electrons: bool = True, expected_charge: int = 0, expected_radicals_count: int = 0,
                 allow_errors: bool = True, logging: bool = False) -> Union[bool, List[str]]:
        """
        Saturate molecules with double and triple bonds and charges and radical states to correct valences of atoms.
        Note: works only with fully explicit hydrogens!

        :param neighbors_distances: If given longest bonds can be removed if need.
        :param reset_electrons: Can change charges and radicals if need.
        :param expected_charge: Reset charge to given. Works only with reset_electrons=True.
        :param expected_radicals_count: Reset radical atoms count to given. Works only with reset_electrons=True.
        :param allow_errors: allow unbalanced result.
        :param logging: return log.
        """
        if any(b != 1 for _, _, b in self.bonds()):
            raise ValenceError('only single bonded skeleton can be saturated')
        atoms = self._atoms
        if not reset_electrons:
            expected_radicals_count = any(self._radicals.values())
            expected_charge = int(self)

        sat, adjacency = _find_possible_valences(atoms, neighbors_distances or self._bonds,
                                                 {x: None for x in self._atoms} if reset_electrons else self._charges,
                                                 {x: None for x in self._atoms} if reset_electrons else self._radicals,
                                                 neighbors_distances is not None)
        charges = {}  # new charge states
        radicals = {}  # new radical states
        bonds = {n: {} for n in atoms}  # new bonds

        seen = set()
        unsaturated = {}
        for n, env in adjacency.items():  # set single bonds in molecule. collect unsaturated atoms
            s = sat[n]
            if len(s) == 1:
                c, r, h = s.pop()
                if not h:
                    seen.add(n)
                    charges[n] = c
                    radicals[n] = r
                    for m in env:
                        if m not in seen:
                            bonds[n][m] = bonds[m][n] = b = Bond(1)
                            b._attach_graph(self, n, m)
                else:
                    unsaturated[n] = [(c, r, h)]
            else:
                # radicals have the lowest priority
                # tuned priority
                # multiple bonds have higher priority
                # charges priority: 0>-1>1>2>3>-2>4>-3>-4
                unsaturated[n] = sorted(s, key=lambda x: (x[1],
                                                          tuned_priority.get((atoms[n].atomic_number, x[0], x[2]), 0),
                                                          -x[2], charge_priority[x[0]]))

        log = []
        if (need_radicals := expected_radicals_count - sum(radicals.values())) < 0:
            log.append('Radical state not balanced')
            if not allow_errors:
                if logging:
                    return log
                return False
            need_radicals = 0  # reset to zero

        if unsaturated:
            # create graph of unsaturated atoms
            bonds_graph = {n: {m for m in adjacency[n] if m in unsaturated} for n in unsaturated}
            order = list(unsaturated)
            # try to saturate with different random states
            for _ in range(len(unsaturated)):
                shuffle(order)
                sb, sa, log_ = _saturate({n: bonds_graph[n].copy() for n in order}, unsaturated, need_radicals,
                                         expected_charge - sum(charges.values()))
                if not log_:  # success
                    break
            else:  # failed
                if log_ == 1:
                    log.append('Charge state not balanced')
                elif log_ == 2:
                    log.append('Radical state not balanced')
                else:
                    log.append('Charge state not balanced')
                    log.append('Radical state not balanced')
                if not allow_errors:  # all attempts failed
                    if logging:
                        return log
                    return False

            for n, m, b in sb:
                bonds[n][m] = bonds[m][n] = b = Bond(b)
                b._attach_graph(self, n, m)
            for n, c, r in sa:
                charges[n] = c
                radicals[n] = r
        elif expected_charge != sum(charges.values()):  # check charge for saturated case
            log.append('Charge state not balanced')
            if not allow_errors:
                if logging:
                    return log
                return False
        # reset molecule
        self._bonds = bonds
        self._radicals = radicals
        self._charges = charges
        self._hydrogens = {x: 0 for x in atoms}  # reset invalid hydrogens counts.
        self.flush_cache()
        if logging:
            if not log:  # check for errors
                log.append('Saturated successfully')
            else:
                log.append('Saturated with errors')
            return log
        return True


def _find_possible_valences(atoms, neighbors_distances, charges, radicals, allow_deleting=True):
    if allow_deleting:
        possible_bonds = {n: md.copy() for n, md in neighbors_distances.items()}
    else:
        possible_bonds = {n: list(md) for n, md in neighbors_distances.items()}
    while True:
        saturation = defaultdict(set)
        for n, env in possible_bonds.items():
            env_atoms = None
            el = len(env)
            dc = charges[n]
            dr = radicals[n]
            for charge, is_radical, valence, implicit, explicit_dict in atoms[n]._compiled_saturation_rules:
                if valence < el or dc is not None and dc != charge or dr is not None and dr != is_radical:
                    continue  # skip impossible rules
                if explicit_dict:
                    if env_atoms is None:  # lazy caching
                        env_atoms = defaultdict(int)
                        for m in env:
                            env_atoms[atoms[m].atomic_number] += 1
                    env_atoms_copy = env_atoms.copy()
                    for (b, a), c in explicit_dict.items():  # stage 1. find explicit valence
                        if env_atoms_copy[a] < c:  # `c` always > 0
                            break  # rule not matched
                        env_atoms_copy[a] -= c
                    else:  # stage 2. find possible valence
                        if unmatched := sum(env_atoms_copy.values()):  # number of atoms outside rule
                            if implicit >= unmatched:
                                # number of implicit H should be greater or equal to number of neighbors
                                saturation[n].add((charge, is_radical, valence - el))
                        else:  # pattern fully matched. difference bw valence and connectivity is unsaturation.
                            saturation[n].add((charge, is_radical, valence - el))
                else:   # unspecific rule. found possible valence
                    saturation[n].add((charge, is_radical, valence - el))
            if n not in saturation:  # valence not found
                break
        else:  # all atoms passed
            break
        if allow_deleting:
            out = max(env.items(), key=itemgetter(1))[0]
            del possible_bonds[out][n]
            del possible_bonds[n][out]
        else:
            raise ValenceError('Structure has invalid atoms neighbors count and electron states')
    return saturation, possible_bonds


def _saturate(bonds, atoms, expected_radicals_count, expected_charge):
    atoms = {k: v.copy() for k, v in atoms.items()}
    dots = []
    saturation = []
    electrons = []
    while True:
        # get isolated atoms. atoms should be charged or radical
        to_del = []
        for n, env in bonds.items():
            if not env:
                es = [(n, c, r) for c, r, h in atoms[n] if not h]
                if not es:
                    raise ValenceError('Saturation impossible. '
                                       f"Isolated atom ({n}) doesn't have appropriate charge-radical state")
                to_del.append(n)
                dots.append(es)
        for n in to_del:
            del bonds[n]
        if not bonds:
            break

        try:  # get terminal atom
            n = next(n for n, ms in bonds.items() if len(ms) == 1)
        except StopIteration:
            # get ring or linker atom
            n, _ = min(bonds.items(), key=lambda x: len(x[1]))
            m = bonds[n].pop()
            bonds[m].discard(n)

            for (nc, nr, nh), (i, (mc, mr, mh)) in product(atoms[n], enumerate(atoms[m])):
                if nh == mh:
                    saturation.append((n, m, nh + 1))
                    electrons.append((n, nc, nr))
                    electrons.append((m, mc, mr))

                    for x in bonds.pop(n):
                        saturation.append((n, x, 1))
                        bonds[x].discard(n)
                    for x in bonds.pop(m):
                        saturation.append((m, x, 1))
                        bonds[x].discard(m)
                    break
                elif nh < mh:
                    electrons.append((n, nc, nr))
                    saturation.append((n, m, nh + 1))
                    atoms[m].pop(i)
                    atoms[m].insert(i, (mc, mr, mh - nh))

                    for x in bonds.pop(n):
                        saturation.append((n, x, 1))
                        bonds[x].discard(n)
                    break
                elif nh > mh:
                    electrons.append((m, mc, mr))
                    saturation.append((n, m, mh + 1))
                    atoms[n].pop(i)
                    atoms[n].insert(i, (nc, nr, nh - mh))

                    for x in bonds.pop(m):
                        saturation.append((m, x, 1))
                        bonds[x].discard(m)
                    break
        else:
            m = bonds.pop(n).pop()
            bonds[m].discard(n)

            for (nc, nr, nh), (i, (mc, mr, mh)) in product(atoms[n], enumerate(atoms[m])):
                if nh == mh:
                    saturation.append((n, m, nh + 1))
                    electrons.append((n, nc, nr))
                    electrons.append((m, mc, mr))
                    for x in bonds.pop(m):
                        saturation.append((m, x, 1))
                        bonds[x].discard(m)
                    break
                elif nh < mh and bonds[m]:
                    electrons.append((n, nc, nr))
                    saturation.append((n, m, nh + 1))
                    atoms[m].pop(i)
                    atoms[m].insert(i, (mc, mr, mh - nh))
                    break
            else:
                saturation.append((n, m, 1))
                if not bonds[m]:
                    del bonds[m]

    combo_ua = []  # possible single atoms electron states
    for s in dots:
        if len(s) == 1:
            electrons.extend(s)
        elif s:
            combo_ua.append(s)

    # if < 0 - we already in bad situation
    # if > 0 - we need more radicals
    need_radical = expected_radicals_count - sum(x for _, _, x in electrons)
    need_charge = expected_charge - sum(x for _, x, _ in electrons)
    if combo_ua:
        # try randomly set charges and radicals.
        # first pick required radical states.
        # second try to minimize charge delta.
        for attempt in range(1, len(combo_ua) + 1):
            shuffle(combo_ua)
            charges_radicals = []
            rad = []
            chg = []
            for atom in combo_ua:
                if len(rad) < need_radical:  # pick radicals
                    r = next((x for x in atom if x[2]), None)
                    if r:  # pick random radical states
                        rad.append(r)
                    else:  # not radical
                        chg.append(atom)
                else:  # pick not radical states
                    c = [x for x in atom if not x[2]]
                    if len(c) > 1:
                        chg.append(c)
                    elif c:
                        charges_radicals.extend(c)
                    elif attempt == len(combo_ua):  # all states has radical. balancing impossible
                        chg.append(atom)  # fuck it horse. we in last attempt
                    else:  # do next attempt
                        break
            else:
                charges_radicals.extend(rad)
                current_charge = need_charge - sum(x for _, x, _ in charges_radicals)
                current_radical = need_radical - len(rad)
                for x in chg:
                    n, c, r = min(x, key=lambda x: abs(current_charge - x[1]))
                    charges_radicals.append((n, c, r))
                    current_charge -= c
                    current_radical -= r

                if current_radical:  # radical unbalanced
                    if current_charge:
                        log = 3
                    else:
                        log = 2
                elif current_charge:
                    log = 1
                else:  # balanced!
                    log = 0
                    break

        electrons.extend(charges_radicals)
    elif need_radical:
        log = 3 if need_charge else 2
    elif need_charge:
        log = 1
    else:
        log = 0
    return saturation, electrons, log


__all__ = ['Saturation']



# -*- coding: utf-8 -*-
#
#  Copyright 2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from re import compile, fullmatch, match, search
from .._mdl import common_isotopes
from ...containers.bonds import QueryBond
from ...exceptions import IncorrectSmiles, IncorrectSmarts
from ...periodictable.element import ListElement


# -,= OR bonds supported
# !: NOT bonds supported
# ~ ANY bonds supported
# @ bond not supported
# @;!: and any other complicated combinations not supported


# tokens structure:
# (type: int, value)
# types:
# 0: atom
# 1: bond
# 2: open chain (
# 3: close chain )
# 4: dot bond .
# 5: in bracket raw data []
# 6: closure number
# 7: raw closure number
# 8: aromatic atom
# 9: up down bond

# 10: query OR bond
# 11: query NOT bond
# 12: in ring bond


atomic_numbers = dict(enumerate(common_isotopes, 1))
iso_re = compile(r'^[0-9]+')
chg_re = compile(r'[+-][1-4+-]?')
mpp_re = compile(r':[1-9][0-9]*$')
str_re = compile(r'@[@?]?')

replace_dict = {'-': 1, '=': 2, '#': 3, ':': 4, '~': 8}
not_dict = {'-': [2, 3, 4], '=': [1, 3, 4], '#': [1, 2, 4], ':': [1, 2, 3]}
atom_re = compile(r'([1-9][0-9]{0,2})?([A-IK-PR-Zacnopsbt][a-ik-pr-vy]?)(@@|@)?(H[1-4]?)?([+-][1-4+-]?)?(:[0-9]{1,4})?')
charge_dict = {'+': 1, '+1': 1, '++': 2, '+2': 2, '+3': 3, '+++': 3, '+4': 4, '++++': 4,
               '-': -1, '-1': -1, '--': -2, '-2': -2, '-3': -3, '---': -3, '-4': -4, '----': -4}


def _tokenize(smiles):
    token_type = token = None
    tokens = []
    for s in smiles:
        if token_type == 12:  # -;, =;, #; or :; found.
            if s == '!':  # not in ring
                if token:  # !! case
                    raise IncorrectSmarts('Invalid ring bond token')
                token = True
            elif s == '@':
                tokens.append((12, QueryBond(tokens.pop(-1)[1], not token)))
                token_type = token = None  # finalize [not]ring bond
            else:
                raise IncorrectSmarts('Invalid ring bond token')
        # [atom block parser]
        elif s == '[':  # open complex token
            if token_type == 5:  # two opened [
                raise IncorrectSmiles('[..[')
            elif token_type in (10, 11):
                raise IncorrectSmarts('Query bond invalid')
            elif token_type == 7:  # empty closure
                raise IncorrectSmiles('invalid closure')
            elif token:
                tokens.append((token_type, token))
            token = []
            token_type = 5
        elif s == ']':  # close complex token
            if token_type != 5:
                raise IncorrectSmiles(']..]')
            elif not token:
                raise IncorrectSmiles('empty [] brackets')
            tokens.append((5, ''.join(token)))
            token = None
            token_type = 0  # mark as atom
        elif token_type == 5:  # grow token with brackets. skip validation
            token.append(s)
        # closure parser
        elif s.isnumeric():  # closures
            if token_type in (10, 11):
                raise IncorrectSmarts('Query bond invalid')
            elif token_type == 2:
                raise IncorrectSmiles('(1 case invalid')
            elif token_type == 7:  # % already found. collect number
                if not token and s == '0':
                    raise IncorrectSmiles('number starts with 0')
                token.append(s)
                if len(token) == 2:
                    tokens.append((6, int(''.join(token))))
                    token = None
                    token_type = 6  # mark finished
            else:
                if s == '0':
                    raise IncorrectSmiles('number starts with 0')
                elif token:
                    tokens.append((token_type, token))
                    token = None
                token_type = 6
                tokens.append((6, int(s)))
        elif token_type == 7:
            raise IncorrectSmiles('expected closure number')
        elif s == '%':
            if token_type in (10, 11):
                raise IncorrectSmarts('Query bond invalid')
            elif token_type == 2:
                raise IncorrectSmiles('(%10 case invalid')
            elif token:
                tokens.append((token_type, token))
            token_type = 7
            token = []
        # bonds parser
        elif s in '=#:-~':  # bonds found
            if token_type == 10:
                token.append(replace_dict[s])
                tokens.append((10, token))
                token_type = token = None  # finalize token
            elif token_type == 11:
                token_type = None
                tokens.append((10, not_dict[s]))
            else:
                if token:
                    tokens.append((token_type, token))
                    token = None
                token_type = 1
                tokens.append((1, replace_dict[s]))
        elif token_type in (10, 11):  # expected bond symbol
            raise IncorrectSmarts('query bond invalid')
        elif s in r'\/':
            if token:
                tokens.append((token_type, token))
                token = None
            token_type = 9
            tokens.append((9, s == '/'))  # Up is true
        elif s == '.':
            if token:
                tokens.append((token_type, token))
                token = None
            token_type = 4
            tokens.append((4, None))
        elif s == ';':  # ;@ or ;!@ - in ring or not in ring bond
            if token_type is not None and token_type != 1:  # start of smiles or bond, list of bonds and not bond.
                raise IncorrectSmarts('Ring bond token invalid')
            token_type = 12
        elif s == ',':  # query bond separator
            if token_type != 1:
                raise IncorrectSmarts('Query bond invalid')
            token_type = 10
            token = [tokens.pop(-1)[1]]
        elif s == '!':  # query not bond
            if token_type not in (0, 2, 3, 6, 8):  # closures, brackets or atoms expected
                raise IncorrectSmarts('Query bond invalid')
            elif token:
                tokens.append((token_type, token))
                token = None
            token_type = 11
        # brackets
        elif s == '(':
            if token_type == 2:  # barely opened
                raise IncorrectSmiles('((')
            elif token:
                tokens.append((token_type, token))
                token = None
            token_type = 2
            tokens.append((2, None))
        elif s == ')':
            if token_type == 2:  # barely opened
                raise IncorrectSmiles('()')
            elif token:
                tokens.append((token_type, token))
                token = None
            token_type = 3
            tokens.append((3, None))
        # simple atoms
        elif s in 'NOPSFI':  # organic atoms
            if token:
                tokens.append((token_type, token))
                token = None
            token_type = 0
            tokens.append((0, s))
        elif s in 'cnopsb':  # aromatic ring atom
            if token:
                tokens.append((token_type, token))
                token = None
            token_type = 8
            tokens.append((8, s.upper()))
        elif s in 'CB':  # flag possible Cl or Br
            if token:
                tokens.append((token_type, token))
            token_type = 0
            token = s
        elif token_type == 0:
            if s == 'l':
                if token == 'C':
                    tokens.append((0, 'Cl'))
                    token = None
                else:
                    raise IncorrectSmiles('invalid element Bl')
            elif s == 'r':
                if token == 'B':
                    tokens.append((0, 'Br'))
                    token = None
                else:
                    raise IncorrectSmiles('invalid smiles for Cr')
            else:
                raise IncorrectSmiles('invalid smiles')
        else:
            raise IncorrectSmiles('invalid smiles')

    if token_type == 5:
        raise IncorrectSmiles('atom description has not finished')
    elif token_type == 7:
        if token:
            tokens.append((6, int(token[0])))
        else:
            raise IncorrectSmiles('invalid %closure')
    elif token:
        tokens.append((token_type, token))  # C or B
    return tokens


def _atom_parse(token):
    # [isotope]Element[element][@[@]][H[n]][+-charge][:mapping]
    match = fullmatch(atom_re, token)
    if match is None:
        raise IncorrectSmiles(f'atom token invalid {{{token}}}')
    isotope, element, stereo, hydrogen, charge, mapping = match.groups()

    if isotope:
        isotope = int(isotope)

    if stereo:
        stereo = stereo == '@'

    if hydrogen:
        if len(hydrogen) > 1:
            hydrogen = int(hydrogen[1:])
        else:
            hydrogen = 1
    else:
        hydrogen = 0

    if charge:
        try:
            charge = charge_dict[charge]
        except KeyError:
            raise IncorrectSmiles('charge token invalid')
    else:
        charge = 0

    if mapping:
        try:
            mapping = int(mapping[1:])
        except ValueError:
            raise IncorrectSmiles('invalid mapping token')
    else:
        mapping = 0

    if element in ('c', 'n', 'o', 'p', 's', 'as', 'se', 'b', 'te'):
        _type = 8
        element = element.capitalize()
    else:
        _type = 0
    return _type, {'element': element, 'isotope': isotope, 'mapping': mapping, 'charge': charge, 'is_radical': False,
                   'x': 0., 'y': 0., 'z': 0., 'hydrogen': hydrogen, 'stereo': stereo}


def _query_parse(token):
    if isotope := match(iso_re, token):
        token = token[isotope.end():]  # remove isotope substring
        isotope = int(isotope.group())
    if charge := search(chg_re, token):
        token = token[:charge.start()] + token[charge.end():]  # remove charge substring
        charge = charge_dict[charge.group()]
    else:
        charge = 0
    if mapping := search(mpp_re, token):
        token = token[:mapping.start()]
        mapping = int(mapping.group()[1:])
    else:
        mapping = 0
    if stereo := search(str_re, token):  # drop stereo mark. unsupported
        token = token[:stereo.start()] + token[stereo.end():]

    # supported only <;> and <,> logic. <&> and silent <&> not supported!
    primitives = token.split(';')
    if element := primitives[0]:
        element = [int(x[1:]) if x.startswith('#') else x for x in element.split(',')]
        if len(element) == 1:
            element = element[0]
        else:  # only atoms supported
            tmp = []
            for x in element:
                if isinstance(x, int):
                    try:
                        tmp.append(atomic_numbers[x])
                    except KeyError as e:
                        raise IncorrectSmiles('Invalid atomic number') from e
                elif x in common_isotopes:
                    tmp.append(x)
                else:
                    raise IncorrectSmarts('Invalid element symbol')
            element = ListElement(tmp)
    else:
        raise IncorrectSmarts('Empty element')

    hybridization = rings_sizes = neighbors = hydrogens = heteroatoms = None
    masked = False
    for p in primitives[1:]:  # parse hydrogens (h), neighbors (D), rings_sizes (r or !R), hybridization == 4 (a)
        if not p:
            continue
        elif p == 'a':  # aromatic atom
            hybridization = 4
        elif p == 'A':  # ignore aliphatic mark. Ad-Hoc for Marwin.
            continue
        elif p == '!R':
            rings_sizes = 0
        elif p == 'M':
            masked = True
        else:
            p = p.split(',')
            if len(p) != 1 and len({x[0] for x in p}) > 1:
                raise IncorrectSmarts('Unsupported OR statement')
            elif (t := p[0][0]) not in ('D', 'h', 'r', 'x', 'z'):
                raise IncorrectSmarts('Unsupported SMARTS primitive. Use only D, h, r, !R and a.')
                # z and x private chython marks for hybridization and heteroatoms count
            try:
                p = [int(x[1:]) for x in p]
            except ValueError:
                raise IncorrectSmarts('Unsupported SMARTS primitive')

            if t == 'D':
                neighbors = p
            elif t == 'h':
                hydrogens = p
            elif t == 'r':  # r
                rings_sizes = p
            elif t == 'x':
                heteroatoms = p
            else:  # z
                hybridization = p

    return 0, {'element': element, 'isotope': isotope, 'mapping': mapping, 'charge': charge, 'is_radical': False,
               'heteroatoms': heteroatoms, 'hydrogens': hydrogens, 'neighbors': neighbors,
               'rings_sizes': rings_sizes, 'hybridization': hybridization, 'masked': masked}


def smiles_tokenize(smi):
    tokens = _tokenize(smi)
    out = []
    for token_type, token in tokens:
        if token_type in (0, 8):  # simple atom
            out.append((token_type, {'element': token, 'isotope': None, 'mapping': 0, 'charge': 0, 'is_radical': False,
                                     'x': 0., 'y': 0., 'z': 0., 'hydrogen': None, 'stereo': None}))
        elif token_type == 5:
            out.append(_atom_parse(token))
        elif token_type == 10:
            raise IncorrectSmiles('SMARTS detected')
        else:
            out.append((token_type, token))
    return out


def smarts_tokenize(smi):
    tokens = _tokenize(smi)
    out = []
    for token_type, token in tokens:
        if token_type in (0, 8):  # simple atom
            out.append((0, {'element': token, 'isotope': None, 'mapping': 0, 'charge': 0, 'is_radical': False,
                            'heteroatoms': None, 'hydrogens': None, 'neighbors': None,
                            'rings_sizes': None, 'hybridization': None, 'masked': False}))
        elif token_type == 5:
            out.append(_query_parse(token))
        else:
            out.append((token_type, token))
    return out


__all__ = ['smiles_tokenize', 'smarts_tokenize']



# -*- coding: utf-8 -*-
#
#  Copyright 2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from itertools import combinations, product
from typing import TYPE_CHECKING, Union, List
from ._acid import rules as acid_rules, stripped_rules as stripped_acid_rules
from ._base import rules as base_rules, stripped_rules as stripped_base_rules


if TYPE_CHECKING:
    from chython import MoleculeContainer


class AcidBase:
    __slots__ = ()

    def neutralize(self: 'MoleculeContainer', *, keep_charge=True, logging=False,
                   _fix_stereo=True) -> Union[bool, List[int]]:
        """
        Convert organic salts to neutral form if possible. Only one possible form used for charge unbalanced structures.

        :param keep_charge: do partial neutralization to keep total charge of molecule.
        :param logging: return changed atoms list.
        """
        try:
            mol, changed = next(self._neutralize(keep_charge))
        except StopIteration:
            if logging:
                return []
            return False

        self._charges.update(mol._charges)
        self._hydrogens.update(mol._hydrogens)
        self.flush_cache()
        if _fix_stereo:
            self.fix_stereo()
        if logging:
            return list(changed)
        return True

    def enumerate_charged_forms(self: 'MoleculeContainer', *, deep: int = 4, limit: int = 1000):
        """
        Enumerate protonated and deprotonated ions. Use on neutralized molecules.

        :param deep: Maximum amount of added or removed protons.
        :param limit: Maximum amount of generated structures.
        """
        if limit < 1:
            raise ValueError('limit should be greater or equal 1')

        donors = set()
        acceptors = set()
        for q in acid_rules:
            for mapping in q.get_mapping(self, automorphism_filter=False):
                donors.add(mapping[1])
        for q in base_rules:
            for mapping in q.get_mapping(self, automorphism_filter=False):
                acceptors.add(mapping[1])

        h_source = [None] * deep + list(donors)
        h_drain = list(acceptors) + [None] * deep

        seen = set()
        seen_combo = set()
        for r in range(min(len(acceptors), deep), 0, -1):  # number of acceptors to protonate
            for ac in combinations(acceptors, r):  # acceptors atoms to protonate
                uniq = set()
                for dc in combinations(h_source, r):  # H donors
                    if dc in uniq:  # filter None-containing duplicates
                        continue
                    uniq.add(dc)
                    seen_combo.add((dc, ac))
                    mol = self.copy()
                    for n in ac:
                        mol._hydrogens[n] += 1
                        mol._charges[n] += 1
                    for n in dc:
                        if n is not None:
                            mol._hydrogens[n] -= 1
                            mol._charges[n] -= 1
                    if mol not in seen:
                        seen.add(mol)
                        yield mol
                    limit -= 1
                    if not limit:
                        return

        for r in range(1, min(len(donors), deep) + 1):
            for dc in combinations(donors, r):
                uniq = set()
                for ac in combinations(h_drain, r):  # H acceptors
                    if ac in uniq:
                        continue
                    uniq.add(ac)
                    if (dc, ac) in seen_combo:
                        continue
                    mol = self.copy()
                    for n in ac:
                        if n is not None:
                            mol._hydrogens[n] += 1
                            mol._charges[n] += 1
                    for n in dc:
                        if n is not None:
                            mol._hydrogens[n] -= 1
                            mol._charges[n] -= 1
                    if mol not in seen:
                        seen.add(mol)
                        yield mol
                    limit -= 1
                    if not limit:
                        return

    def _neutralize(self: 'MoleculeContainer', keep_charge=True):
        donors = set()
        acceptors = set()
        for q in stripped_acid_rules:
            for mapping in q.get_mapping(self, automorphism_filter=False):
                donors.add(mapping[1])
        for q in stripped_base_rules:
            for mapping in q.get_mapping(self, automorphism_filter=False):
                acceptors.add(mapping[1])

        if keep_charge:
            if not donors or not acceptors:
                return  # neutralization impossible
            elif len(donors) > len(acceptors):
                copy = self.copy()
                for a in acceptors:
                    copy._hydrogens[a] += 1
                    copy._charges[a] += 1
                for c in combinations(donors, len(acceptors)):
                    mol = copy.copy()
                    for d in c:
                        mol._hydrogens[d] -= 1
                        mol._charges[d] -= 1
                    yield mol, acceptors.union(c)
            elif len(donors) < len(acceptors):
                copy = self.copy()
                for d in donors:
                    copy._hydrogens[d] -= 1
                    copy._charges[d] -= 1
                for c in combinations(acceptors, len(donors)):
                    mol = copy.copy()
                    for a in c:
                        mol._hydrogens[a] += 1
                        mol._charges[a] += 1
                    yield mol, donors.union(c)
            else:  # balanced!
                mol = self.copy()
                for d in donors:
                    mol._hydrogens[d] -= 1
                    mol._charges[d] -= 1
                for a in acceptors:
                    mol._hydrogens[a] += 1
                    mol._charges[a] += 1
                yield mol, donors | acceptors
        elif donors or acceptors:
            mol = self.copy()
            for d in donors:
                mol._hydrogens[d] -= 1
                mol._charges[d] -= 1
            for a in acceptors:
                mol._hydrogens[a] += 1
                mol._charges[a] += 1
            yield mol, donors | acceptors

    def _enumerate_zwitter_tautomers(self: 'MoleculeContainer'):
        donors = set()
        acceptors = set()
        for q in acid_rules:
            for mapping in q.get_mapping(self, automorphism_filter=False):
                donors.add(mapping[1])
        for q in base_rules:
            for mapping in q.get_mapping(self, automorphism_filter=False):
                acceptors.add(mapping[1])

        for d, a in product(donors, acceptors):
            mol = self.copy()
            mol._hydrogens[d] -= 1
            mol._hydrogens[a] += 1
            mol._charges[d] -= 1
            mol._charges[a] += 1
            yield mol


__all__ = ['AcidBase']



# -*- coding: utf-8 -*-
#
#  Copyright 2017-2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#


class EmptyMolecule(ValueError):
    """
    Molecule without atoms
    """


class MappingError(ValueError):
    """
    Atom-to-Atom mapping invalid
    """


class AtomNotFound(KeyError):
    """
    Bad atom number
    """


class BondNotFound(KeyError):
    """
    Bad atoms numbers
    """


class NotChiral(KeyError):
    """
    Atom not chiral
    """


class IsChiral(KeyError):
    """
    Atom already chiral
    """


class InvalidAromaticRing(ValueError):
    """
    Aromatic ring has impossible Kekule structure
    """


class IsConnectedAtom(Exception):
    """
    Atom is already attached to graph
    """


class IsNotConnectedAtom(Exception):
    """
    Atom is not attached to graph
    """


class IsConnectedBond(Exception):
    """
    Bond is already attached to graph
    """


class IsNotConnectedBond(Exception):
    """
    Bond is not attached to graph
    """


class ValenceError(Exception):
    """
    Atom has valence error
    """


class IncorrectSmiles(ValueError):
    """
    SMILES string invalid
    """


class IncorrectSmarts(IncorrectSmiles):
    """
    SMARTS string invalid or unsupported
    """


class ImplementationError(Exception):
    """
    Algorithm has errors. Please send example of structure to author for analyze.
    """


class ParseError(ValueError):
    """
    File parsing error
    """
    __slots__ = ('number', 'position', 'log', 'meta', 'structure', 'errors')

    def __init__(self, number, position, log, meta, structure=None, errors=None):
        super().__init__()
        self.number = number
        self.position = position
        self.log = log
        self.meta = meta
        self.structure = structure
        self.errors = errors


class EmptyV2000(ValueError):
    """Empty V2000 in EMol parser"""


class ParseReactionError(ValueError):
    __slots__ = ('structure', 'errors')

    def __init__(self, structure, errors):
        self.structure = structure
        self.errors = errors



# -*- coding: utf-8 -*-
#
#  Copyright 2018-2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2021 Timur Gimadiev <timur.gimadiev@gmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from collections import defaultdict
from typing import List, Tuple, TYPE_CHECKING, Union
from ._reagents import *
from ...exceptions import MappingError


if TYPE_CHECKING:
    from chython import ReactionContainer


class StandardizeReaction:
    __slots__ = ()

    def canonicalize(self: 'ReactionContainer', *, fix_mapping: bool = True, logging=False, fix_tautomers=True) -> \
            Union[bool, List[Tuple[int, Tuple[int, ...], int, str]]]:
        """
        Convert molecules to canonical forms of functional groups and aromatic rings without explicit hydrogens.
        Return True if in any molecule found not canonical group.

        :param fix_mapping: Search AAM errors of functional groups.
        :param logging: return log from molecules with index of molecule.
            Otherwise, return True if these groups found in any molecule.
        :param fix_tautomers: convert tautomers to canonical forms.
        """
        total = []
        for n, m in enumerate(self.molecules()):
            total.extend((n, *x) for x in m.canonicalize(logging=True, fix_tautomers=fix_tautomers))

        if fix_mapping:
            total.extend((-1, x, -1, m) for m, x in self.fix_groups_mapping(logging=True))

        if total:
            self.flush_cache()
        if logging:
            return total
        return bool(total)

    def standardize(self: 'ReactionContainer', *, fix_mapping: bool = True, logging=False, fix_tautomers=True) -> \
            Union[bool, List[Tuple[int, Tuple[int, ...], int, str]]]:
        """
        Fix functional groups representation.
        Return True if in any molecule fixed group.

        Deprecated method. Use `canonicalize` directly.

        :param fix_mapping: Search AAM errors of functional groups.
        :param logging: return log from molecules with index of molecule.
            Otherwise, return True if these groups found in any molecule.
        :param fix_tautomers: convert tautomers to canonical forms.
        """
        total = []
        for n, m in enumerate(self.molecules()):
            total.extend((n, *x) for x in m.standardize(logging=True, fix_tautomers=fix_tautomers))

        if fix_mapping:
            total.extend((-1, x, -1, m) for m, x in self.fix_groups_mapping(logging=True))

        if total:
            self.flush_cache()
        if logging:
            return total
        return bool(total)

    def thiele(self: 'ReactionContainer', *, fix_tautomers=True) -> bool:
        """
        Convert structures to aromatic form.
        Return True if in any molecule found kekule ring

        :param fix_tautomers: convert tautomers to canonical forms.
        """
        total = False
        for m in self.molecules():
            if m.thiele(fix_tautomers=fix_tautomers) and not total:
                total = True
        if total:
            self.flush_cache()
        return total

    def kekule(self: 'ReactionContainer', *, buffer_size=7) -> bool:
        """
        Convert structures to kekule form.
        Return True if in any molecule found aromatic ring

        :param buffer_size: number of attempts of pyridine form searching.
        """
        total = False
        for m in self.molecules():
            if m.kekule(buffer_size=buffer_size) and not total:
                total = True
        if total:
            self.flush_cache()
        return total

    def clean_isotopes(self: 'ReactionContainer') -> bool:
        """
        Clean isotope marks for all molecules in reaction.
        Returns True if in any molecule found isotope.
        """
        flag = False
        for m in self.molecules():
            if m.clean_isotopes() and not flag:
                flag = True
        if flag:
            self.flush_cache()
        return flag

    def clean_stereo(self: 'ReactionContainer'):
        """
        Remove stereo data
        """
        for m in self.molecules():
            m.clean_stereo()
        self.flush_cache()

    def check_valence(self: 'ReactionContainer') -> List[Tuple[int, Tuple[int, ...]]]:
        """
        Check valences of all atoms of all molecules.

        Works only on molecules with aromatic rings in Kekule form.
        :return: list of invalid molecules with invalid atoms lists
        """
        out = []
        for n, m in enumerate(self.molecules()):
            if c := m.check_valence():
                out.append((n, tuple(c)))
        return out

    def implicify_hydrogens(self: 'ReactionContainer') -> int:
        """
        Remove explicit hydrogens if possible.

        :return: number of removed hydrogens.
        """
        total = 0
        for m in self.molecules():
            total += m.implicify_hydrogens()
        if total:
            self.flush_cache()
        return total

    def explicify_hydrogens(self: 'ReactionContainer') -> int:
        """
        Add explicit hydrogens to atoms

        :return: number of added atoms
        """
        total = 0
        start_map = 0
        for m in self.molecules():
            map_ = max(m, default=0)
            if map_ > start_map:
                start_map = map_

        mapping = defaultdict(list)
        for m in self.reactants:
            maps = m.explicify_hydrogens(_return_map=True, start_map=start_map + 1)
            if maps:
                for n, h in maps:
                    mapping[n].append(h)
                start_map = maps[-1][1]
                total += len(maps)

        for m in self.reagents:
            maps = m.explicify_hydrogens(_return_map=True, start_map=start_map + 1)
            if maps:
                start_map = maps[-1][1]
                total += len(maps)

        for m in self.products:
            maps = m.explicify_hydrogens(_return_map=True, start_map=start_map + 1)
            if maps:
                total += len(maps)
                remap = {}
                free = []
                for n, h in maps:
                    if n in mapping and mapping[n]:
                        remap[h] = mapping[n].pop()
                        free.append(h)
                    elif free:
                        remap[h] = start_map = free.pop(0)
                    else:
                        start_map = h
                m.remap(remap)

        if total:
            self.flush_cache()
        return total

    def remove_reagents(self, *, keep_reagents: bool = False, mapping: bool = True) -> bool:
        """
        Place molecules, except reactants, to reagents list. Reagents - molecules which atoms not presented in products.
        Mapping based approach remove molecules without reaction center.
        Rule based approach remove equal molecules in reactants and products, and predefined reactants.

        :param mapping: use atom-to-atom mapping to detect reagents, otherwise use predefined list of common reagents.
        :param keep_reagents: delete reagents if False

        Return True if any reagent found.
        """
        if mapping:
            return self.__remove_reagents_mapping(keep_reagents)
        return self.__remove_reagents_rules(keep_reagents)

    def __remove_reagents_rules(self: 'ReactionContainer', keep_reagents):
        if not self.reactants or not self.products:  # there is no reaction
            return False

        reactants_st1 = []
        products_st1 = []
        reagents_st1 = set(self.reagents)

        # find the same molecules in reactants and products
        for m in self.reactants:
            if m in self.products:
                reagents_st1.add(m)
            else:
                reactants_st1.append(m)
        for m in self.products:
            if m in self.reactants:
                reagents_st1.add(m)
            else:
                products_st1.append(m)
        if not reactants_st1 or not products_st1:
            return False  # keep bad reaction as is

        reactants_st2 = []
        products_st2 = []
        reagents_st2 = reagents_st1.copy()

        # filter out predefined reagents
        for m in reactants_st1:
            if m in reagents_set:
                reagents_st2.add(m)
            else:
                reactants_st2.append(m)
        for m in products_st1:
            if m in reagents_set:
                reagents_st2.add(m)
            else:
                products_st2.append(m)
        if not reactants_st2 or not products_st2:  # reaction contains only simple molecules. roll-back to step 1
            reactants_st2 = reactants_st1
            products_st2 = products_st1
            reagents_st2 = reagents_st1

        if keep_reagents:
            tmp = []
            for m in self.reagents:
                if m in reagents_st2:
                    tmp.append(m)
                    reagents_st2.discard(m)
            tmp.extend(reagents_st2)
            reagents = tuple(tmp)
        else:
            reagents = ()
        self._ReactionContainer__reactants = tuple(reactants_st2)
        self._ReactionContainer__products = tuple(products_st2)
        self._ReactionContainer__reagents = reagents
        self.flush_cache()
        self.fix_positions()
        return True

    def __remove_reagents_mapping(self: 'ReactionContainer', keep_reagents):
        cgr = ~self
        if cgr.center_atoms:
            active = set(cgr.center_atoms)
            reactants = []
            products = []
            reagents = set(self.reagents)
            for i in self.reactants:
                if not active.isdisjoint(i):
                    reactants.append(i)
                else:
                    reagents.add(i)
            for i in self.products:
                if not active.isdisjoint(i):
                    products.append(i)
                else:
                    reagents.add(i)
            if keep_reagents:
                tmp = []
                for m in self.reagents:
                    if m in reagents:
                        tmp.append(m)
                        reagents.discard(m)
                tmp.extend(reagents)
                reagents = tuple(tmp)
            else:
                reagents = ()

            if len(reactants) != len(self.reactants) or len(products) != len(self.products):
                self._ReactionContainer__reactants = tuple(reactants)
                self._ReactionContainer__products = tuple(products)
                self._ReactionContainer__reagents = reagents
                self.flush_cache()
                self.fix_positions()
                return True
            return False
        raise MappingError("Reaction center is absent according to mapping")

    def contract_ions(self: 'ReactionContainer') -> bool:
        """
        Contract ions into salts (Molecules with disconnected components).
        Note: works only for unambiguous cases. e.g. equal anions/cations and different or equal cations/anions.

        Return True if any ions contracted.
        """
        neutral, cations, anions, total = _sift_ions(self.reagents)
        salts = _contract_ions(anions, cations, total)
        if salts:
            neutral.extend(salts)
            self._ReactionContainer__reagents = tuple(neutral)
            changed = True
        else:
            changed = False

        neutral, cations, anions, total = _sift_ions(self.reactants)
        salts = _contract_ions(anions, cations, total)
        if salts:
            anions_order = {frozenset(m): n for n, m in enumerate(anions)}
            cations_order = {frozenset(m): n for n, m in enumerate(cations)}
            neutral.extend(salts)
            self._ReactionContainer__reactants = tuple(neutral)
            changed = True
        else:
            anions_order = cations_order = {}

        neutral, cations, anions, total = _sift_ions(self.products)
        if cations and anions:
            anions.sort(key=lambda x: anions_order.get(frozenset(x), -1))
            cations.sort(key=lambda x: cations_order.get(frozenset(x), -1))
        salts = _contract_ions(anions, cations, total)
        if salts:
            neutral.extend(salts)
            self._ReactionContainer__products = tuple(neutral)
            changed = True

        if changed:
            self.flush_cache()
            self.fix_positions()
            return True
        return False


def _sift_ions(mols):
    anions = []
    cations = []
    neutral = []
    total = 0
    for m in mols:
        c = int(m)
        total += c
        if c > 0:
            cations.append(m)
        elif c < 0:
            anions.append(m)
        else:
            neutral.append(m)
    return neutral, cations, anions, total


def _contract_ions(anions, cations, total):
    if not anions or not cations:  # nothing to contract
        return
    # check ambiguous cases
    if total > 0:
        if len(cations) > 1:  # deficit of anions
            # we have an excess of cations. we can't assign anions univocally
            return  # unite is ambiguous
        salt = cations[0]
        shift_x = salt._fix_plane_mean(0) + 1
        for x in anions:
            shift_x = x._fix_plane_mean(shift_x) + 1
            salt = salt | x
        return [salt]
    elif total < 0:
        if len(anions) > 1:  # deficit of cations
            # we have an excess of anions. we can't assign cations univocally
            return  # unite is ambiguous
        salt = anions[0]
        shift_x = salt._fix_plane_mean(0) + 1
        for x in cations:
            shift_x = x._fix_plane_mean(shift_x) + 1
            salt = salt | x
        return [salt]
    elif len(set(anions)) > 1 and len(set(cations)) > 1:  # different anions and cations
        return

    salts = []
    anions = anions.copy()
    cations = cations.copy()
    while anions:
        ct = cations.pop()
        an = anions.pop()
        shift_x = ct._fix_plane_mean(0) + 1
        shift_x = an._fix_plane_mean(shift_x) + 1
        salt = ct | an
        while True:
            c = int(salt)
            if c > 0:
                an = anions.pop()
                shift_x = an._fix_plane_mean(shift_x) + 1
                salt = salt | an
            elif c < 0:
                ct = cations.pop()
                shift_x = ct._fix_plane_mean(shift_x) + 1
                salt = salt | ct
            else:
                break
        salts.append(salt)
    return salts


__all__ = ['StandardizeReaction']



# -*- coding: utf-8 -*-
#
#  Copyright 2017-2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from functools import cached_property
from itertools import groupby
from logging import getLogger
from operator import itemgetter
from typing import Dict, TYPE_CHECKING


logger = getLogger('chython.morgan')


if TYPE_CHECKING:
    from chython.containers.graph import Graph


class Morgan:
    __slots__ = ()

    @cached_property
    def atoms_order(self: 'Graph') -> Dict[int, int]:
        """
        Morgan like algorithm for graph nodes ordering

        :return: dict of atom-order pairs
        """
        atoms = self._atoms
        if not atoms:  # for empty containers
            return {}
        elif len(atoms) == 1:  # optimize single atom containers
            return dict.fromkeys(atoms, 1)
        ring = self.ring_atoms
        return _morgan({n: hash((hash(a), n in ring)) for n, a in atoms.items()}, self.int_adjacency)

    @cached_property
    def int_adjacency(self: 'Graph') -> Dict[int, Dict[int, int]]:
        """
        Adjacency with integer-coded bonds.
        """
        return {n: {m: hash(b) for m, b in mb.items()} for n, mb in self._bonds.items()}


def _morgan(atoms: Dict[int, int], bonds: Dict[int, Dict[int, int]]) -> Dict[int, int]:
    tries = len(atoms) - 1
    numb = len(set(atoms.values()))
    stab = old_numb = 0

    for _ in range(tries):
        atoms = {n: hash((atoms[n], *(x for x in sorted((atoms[m], b) for m, b in ms.items()) for x in x)))
                 for n, ms in bonds.items()}
        old_numb, numb = numb, len(set(atoms.values()))
        if numb == len(atoms):  # each atom now unique
            break
        elif numb == old_numb:  # not changed. molecules like benzene
            if stab == 3:
                break
            stab += 1
        elif stab:  # changed unique atoms number. reset stability check.
            stab = 0
    else:
        if numb < old_numb:
            logger.warning('number of attempts exceeded. uniqueness has decreased.')

    return {n: i for i, (_, g) in enumerate(groupby(sorted(atoms.items(), key=itemgetter(1)), key=itemgetter(1)),
                                            start=1) for n, _ in g}


__all__ = ['Morgan']
